import scipy as SP
from . import par_lmm_forest as parUtils
import random

def checkMaf(X, maf=None):
    if maf==None:
        maf = 1.0/X.shape[0]
    Xmaf = (X>0).sum(axis=0)
    Iok = (Xmaf>=(maf*X.shape[0]))
    return SP.where(Iok)[0]

def scale_K(K, verbose=False):
    """scale covariance K such that it explains unit variance"""
    c = SP.sum((SP.eye(len(K)) - (1.0 / len(K)) * SP.ones(K.shape)) * SP.array(K))
    scalar = (len(K) - 1) / c
    if verbose:
        print(('Kinship scaled by: %0.4f' % scalar))
    K = scalar * K
    return K

def update_Kernel(unscaled_kernel, X_upd_in, scale=True):
    #filter and scale SNPs
    X_upd = X_upd_in.copy()
    X_upd = X_upd[:,checkMaf(X_upd)]
    X_upd -= X_upd.mean(axis=0)
    X_upd /= X_upd.std(axis=0)
    X_upd = X_upd.T
    #update kernel
    kernel_out = unscaled_kernel.copy()
    kernel_out -= SP.dot(X_upd.T, X_upd)
    if scale:
        return scale_K(kernel_out)
    else:
        return kernel_out

def estimateKernel(X, msample=None, maf=None, scale=True):
    #1. maf filter
    Xpop = X.copy()
    Xpop = Xpop[:,checkMaf(X, maf)]
    #2. sampling of predictors
    if msample != None:
        msample = SP.random.permutation(X.shape[1])[:msample]
        Xpop = Xpop[:,msample]
    Xpop -= Xpop.mean(axis=0)
    Xpop /= Xpop.std(axis=0)
    Xpop = Xpop.copy().T
    Xpop = SP.array(Xpop, dtype='float')
    Kpop  =  SP.dot(Xpop.T,Xpop)
    if scale:
        return scale_K(Kpop)
    else:
        return Kpop

def k_fold_cross_validation(items, k, randomize=True, seed=True):
    if randomize:
        if seed:
            random.seed(10) # make shure we get similar partitions across methods
        items = list(items)
        random.shuffle(items)
    slices = [items[i::k] for i in range(k)]

    for i in range(k):
        validation = slices[i]
        training = [item
                    for s in slices if s is not validation
                    for item in s]
        yield validation

def crossValidationScheme(folds, n):
    validationList = []
    for validation in k_fold_cross_validation(list(range(n)), folds):
        indexes = SP.ones(n) == 0
        indexes[validation] = True
        validationList.append(indexes)
    return validationList

def crossValidate(y, X, K=None, folds=3, model=None, returnModel=False):
    errors =  SP.empty(folds)
    n = y.shape[0]
    indexes = crossValidationScheme(folds,n)
    predictions = SP.empty(y.shape)
    alpha = []
    alphas = []
    msePath = []
    for cvRun in SP.arange(len(indexes)):
        testIndexes = indexes[cvRun]
        yTrain = y[~testIndexes]
        XTrain = X[~testIndexes]
        if K == None:
            model.fit(XTrain, yTrain)
            prediction = SP.reshape(model.predict(X[testIndexes]), (-1,1))
        else: # models having population structure
            KTrain = K[~testIndexes]
            KTrain = KTrain[:,~testIndexes]
            KTest=K[testIndexes]
            KTest=KTest[:,~testIndexes]
            model.reset()
            model.kernel = KTrain #TODO: make nice integration
            model.fit(XTrain, yTrain)
            prediction = SP.reshape(model.predict(X[testIndexes], k=KTest), (-1,1))
        predictions[testIndexes] = prediction
        errors[cvRun] = predictionError(y[testIndexes], prediction)
        print(('prediction error right now is', errors[cvRun]))
        if returnModel:
            alpha.append(model.alpha)
            alphas.append(model.alphas)
            msePath.append(model.mse_path)
    if returnModel:
        return indexes, predictions, errors, alpha, alphas, msePath
    else:
        return indexes, predictions, errors

def predictionError(yTest, yPredict):
    return ((yTest - yPredict)**2).sum()/SP.float_(yTest.shape[0])

def getQuadraticKernel(X, d=.01):
    K = SP.empty((X.shape[0], X.shape[0]))
    for i in SP.arange(X.shape[0]):
        for j in  SP.arange(X.shape[0]):
            K[i,j] = SP.exp(-0.5/d*(X[i]-X[j])**2)
    return scale_K(K)

def generate_linear_data(n_max, n_step, ssv_g, var):
    x = SP.arange(0,n_max,n_step).reshape(-1,1)
    y = SP.zeros_like(x).reshape(-1,1)*0.0
    X = convertToBinaryPredictor(x)
    Xbg = (SP.random.rand(X.shape[0], X.shape[1]) < .5) * 1.0
    weights = var*SP.random.randn(2,1)
    y += X[:,3:4] * weights[0,:]
    Xbg[:,3:4] = X[:,3:4]
    l = X[:,1:2] * X[:,2:3]
    Xbg[:,1:2] = X[:,1:2]
    Xbg[:,2:3] = X[:,2:3]
    y += l * weights[1,:]
    yTr = y.copy()
    ssv_v = 1.0-ssv_g
    if ssv_g > 0.0:
        ldelta = SP.log(ssv_v/SP.float_(ssv_g))
        K = scale_K(getQuadraticKernel(x, d=20))
    else:
        ldelta = None
        K = SP.eye(y.shape[0])
    y += SP.random.multivariate_normal(SP.zeros(K.shape[0]),ssv_g*K+ssv_v*SP.eye(K.shape[0])).reshape(-1,1)
    return Xbg, x, y, yTr, K, ldelta

def convertToBinaryPredictor(x):
    arr = []
    a = 0
    for i in SP.arange(x.size):
        arr.append(bin(x[i,0])[2:])
        l = max(a, bin(x[i,0])[2:].__len__())
    X = SP.zeros((x.size,l))

    for i in SP.arange(x.size):
        head0=l-arr[i].__len__()
        for j in SP.arange(head0):
            X[i,j] = 0
        for j in SP.arange(arr[i].__len__()):
            X[i,head0+j] = SP.int16(arr[i][j])
    return X

# generates data sets to test the continous version of the mixed forest
def lin_data_cont_predictors(n=100, m=1):
    X = SP.random.randn(n,m)
    beta = SP.random.randn(m,1)
    beta[1:]=0
    y = SP.dot(X,beta)
    return X, y



__author__ = 'damienarnol1'

import numpy as np

from limix.core.covar import SQExpCov
from limix.core.covar import ZKZCov
from limix.core.covar import FixedCov
from limix.core.covar import SumCov
from limix.core.gp import GP
from limix.core.mean import MeanBase
from limix.utils.preprocess import covar_rescaling_factor
import sys


def normalise(mat, axis=1):
    mat -= np.reshape(mat.mean(axis=axis), [mat.shape[0], 1])
    mat /= np.reshape(mat.std(axis=axis), [mat.shape[0], 1])
    mat /= np.sqrt(mat.shape[1])
    return mat


def compute_pairwise_interaction(model, expression_file, position_file, output_file,
                                 protein_1, protein_2):
    rm_diag = True

    if model is not 'full' and model is not 'env':
        raise Exception('model not understood. Please specify a model between full and env')

    # read phenotypes data
    with open(expression_file, 'r') as f:
        prot_tmp = f.readline()
    protein_names = prot_tmp.split(' ')
    protein_names[-1] = protein_names[-1][0:-1]  # removing the newline sign at the end of the last protein
    protein_names = np.reshape(protein_names, [len(protein_names), 1])
    phenotypes = np.loadtxt(expression_file, delimiter=' ', skiprows=1)

    # read position data
    X = np.genfromtxt(position_file, delimiter=',')
    if X.shape[0] != phenotypes.shape[0]:
        raise Exception('cell number inconsistent between position and epression levels ')

    # select phenotype and environmental protein
    phen = [i for i in range(0, len(protein_names)) if protein_names[i] == protein_1]
    env_prot_index = [i for i in range(0, len(protein_names)) if protein_names[i] == protein_2]

    N_cells = phenotypes.shape[0]

    # parameters = np.zeros([phenotypes.shape[1], 6])

    # log_lik = np.zeros(phenotypes.shape[1])

    phenotype = phenotypes[:, phen]
    phenotype -= phenotype.mean()
    phenotype /= phenotype.std()
    phenotype = np.reshape(phenotype, [N_cells, 1])

    phenotypes_tmp = phenotypes[:, env_prot_index]
    # phenotypes_tmp = normalise(phenotypes_tmp)

    Kinship = phenotypes_tmp.dot(phenotypes_tmp.transpose())
    # Kinship -= np.linalg.eigvalsh(Kinship).min() * np.eye(N_cells)
    Kinship *= covar_rescaling_factor(Kinship)

    # create different models and print the result including likelihood
    # create all the covariance terms
    direct_cov = FixedCov(Kinship)

    # noise
    noise_cov = FixedCov(np.eye(N_cells))

    # local_noise
    local_noise_cov = SQExpCov(X)
    # local_noise_cov.length = 100
    # local_noise_cov.act_length = False
    # environment effect
    environment_cov = ZKZCov(X, Kinship, rm_diag)

    # mean term
    mean = MeanBase(phenotype)

    #######################################################################
    # defining model
    #######################################################################
    cov = SumCov(noise_cov, local_noise_cov)
    cov = SumCov(cov, environment_cov)
    # if random_start_point:
    #     environment_cov.length = np.random.uniform(10, 300)
    #     environment_cov.scale = np.random.uniform(1, 15)
    #
    # else:
    #     environment_cov.length = 200
    environment_cov.length = 200
    # environment_cov.act_length = False

    if model == 'full':
        cov = SumCov(cov, direct_cov)
    else:
        direct_cov.scale = 0

    # define and optimise GP
    gp = GP(covar=cov, mean=mean)

    try:
        gp.optimize()
    except:
        print('optimisation', str(phen), 'failed')

    # log_lik[phen] = gp.LML()


    # rescale each terms to sample variance one
    # direct cov: unnecessary as fixed covariance rescaled before optimisation
    # local noise covariance
    tmp = covar_rescaling_factor(local_noise_cov.K() / local_noise_cov.scale)
    local_noise_cov.scale /= tmp
    # env effect
    tmp = covar_rescaling_factor(environment_cov.K() / environment_cov.scale ** 2)
    environment_cov.scale = environment_cov.scale ** 2 / tmp

    results = [protein_1,
               protein_2,
               direct_cov.scale,
               noise_cov.scale,
               local_noise_cov.scale,
               local_noise_cov.length,
               environment_cov.scale,
               environment_cov.length]
    results = np.reshape(results, [1, len(results)])

    result_header = 'protein 1' + ' ' + \
                    'protein 2' + ' ' + \
                    'direct_scale' + ' ' + \
                    'noise_scale' + ' ' + \
                    'local_noise_scale' + ' ' + \
                    'local_noise_length' + ' ' + \
                    'environment_scale' + ' ' + \
                    'environment_length'

    with open(output_file, 'w') as f:
        np.savetxt(f,
                   results,
                   delimiter=' ',
                   header=result_header,
                   fmt='%s',
                   comments='')

    # log_lik_file = output_file + '_loglik'
    # with open(log_lik_file, 'w') as f:
    #     np.savetxt(f, log_lik)


##############################################################################################
# running model
##############################################################################################

protein_1 = sys.argv[1]
protein_2 = sys.argv[2]
output_file = sys.argv[3]
image_dir = sys.argv[4]

expression_profile = image_dir+'/'+'expressions.txt'
position_file = image_dir+'/'+'positions.txt'


compute_pairwise_interaction('env', expression_profile, position_file, output_file, protein_1, protein_2)

##############################################################################################
# trial
##############################################################################################
# protein_1 = 'CREB'
# protein_2 = 'pBAD'
# output_file = '/Users/damienarnol1/Documents/local/pro/PhD/random_effects_cluster/tests/Ay10x1/pairwise_res/res.txt'
# image_dir = '/Users/damienarnol1/Documents/local/pro/PhD/random_effects_cluster/tests/Ay10x1/'
#
# expression_profile = image_dir+'/'+'expressions.txt'
# position_file = image_dir+'/'+'positions.txt'
#
#
# compute_pairwise_interaction('env', expression_profile, position_file, output_file, protein_1, protein_2)




import scipy as sp
import scipy.linalg as la
import scipy.stats as st

def standardize(Y,in_place=False):
    """
    standardize Y in a way that is robust to missing values
    in_place: create a copy or carry out inplace opreations?
    """
    if in_place:
        YY = Y
    else:
        YY = Y.copy()
    for i in range(YY.shape[1]):
        Iok = ~SP.isnan(YY[:,i])
        Ym = YY[Iok,i].mean()
        YY[:,i]-=Ym
        Ys = YY[Iok,i].std()
        YY[:,i]/=Ys
    return YY

def covar_rescaling_factor(C):
    """
    Returns the rescaling factor for the Gower normalizion on covariance matrix C
    the rescaled covariance matrix has sample variance of 1
    """
    n = C.shape[0]
    P = sp.eye(n) - sp.ones((n,n))/float(n)
    trPCP = sp.trace(sp.dot(P,sp.dot(C,P)))
    r = (n-1) / trPCP
    return r

def covar_rescaling_factor_efficient(C):
    """
    Returns the rescaling factor for the Gower normalizion on covariance matrix C
    the rescaled covariance matrix has sample variance of 1
    """
    n = C.shape[0]
    P = sp.eye(n) - sp.ones((n,n))/float(n)
    CP = C - C.mean(0)[:, sp.newaxis]
    trPCP = sp.sum(P * CP)
    r = (n-1) / trPCP
    return r

def covar_rescale(C):
    """
    Perform Gower normalizion on covariance matrix C
    the rescaled covariance matrix has sample variance of 1
    """
    sf = covar_rescaling_factor(C)
    return sf * C


def toRanks(A):
    """
    converts the columns of A to ranks
    """
    AA=sp.zeros_like(A)
    for i in range(A.shape[1]):
        AA[:,i] = st.rankdata(A[:,i])
    AA=sp.array(sp.around(AA),dtype="int")-1
    return AA

def gaussianize(Y):
    """
    Gaussianize X: [samples x phenotypes]
    - each phentoype is converted to ranks and transformed back to normal using the inverse CDF
    """
    N,P = Y.shape

    YY=toRanks(Y)
    quantiles=(sp.arange(N)+0.5)/N
    gauss = st.norm.isf(quantiles)
    Y_gauss=sp.zeros((N,P))
    for i in range(P):
        Y_gauss[:,i] = gauss[YY[:,i]]
    Y_gauss *= -1
    return Y_gauss

def rankStandardizeNormal(Y):    
    """
    Gaussianize X: [samples x phenotypes]
    - each phentoype is converted to ranks and transformed back to normal using the inverse CDF
    """
    return gaussianize(Y)

def regressOut(Y, X, return_b=False):
    """
    regresses out X from Y
    """
    Xd = la.pinv(X)
    b = Xd.dot(Y)
    Y_out = Y-X.dot(b)
    if return_b:
        return Y_out, b
    else:
        return Y_out

def remove_dependent_cols(M, tol=1e-6, display=False):
    """
    Returns a matrix where dependent columsn have been removed 
    """
    R = la.qr(M, mode='r')[0][:M.shape[1], :]
    I = (abs(R.diagonal())>tol)
    if sp.any(~I) and display:
        print(('cols ' + str(sp.where(~I)[0]) +
                ' have been removed because linearly dependent on the others'))
        R = M[:,I]
    else:
        R = M.copy()
    return R

def boxcox(X):
    """
    Gaussianize X using the Box-Cox transformation: [samples x phenotypes]

    - each phentoype is brought to a positive schale, by first subtracting the minimum value and adding 1.
    - Then each phenotype transformed by the boxcox transformation
    """
    X_transformed = sp.zeros_like(X)
    maxlog = sp.zeros(X.shape[1])
    for i in range(X.shape[1]):
        i_nan = sp.isnan(X[:,i])
        values = X[~i_nan,i]
        X_transformed[i_nan,i] = X[i_nan,i]
        X_transformed[~i_nan,i], maxlog[i] = st.boxcox(values-values.min()+1.0)
    return X_transformed, maxlog
    




import sys
import scipy as sp
import numpy as np
import scipy.linalg as LA
import copy
import pdb

from .mean_base import MeanBase
from limix.utils.preprocess import regressOut
from limix.utils.util_functions import to_list
from hcache import Cached, cached
from limix.core.type.observed import *
from limix.core.utils import assert_make_float_array
from limix.core.utils import assert_type_or_list_type
from limix.utils.util_functions import vec


class MeanKronSum(MeanBase):
    """
    Sum of Kronecker Mean for multi-trait gp regression
    Notation:
        N = number of individuals
        No = number of out-of-sample individuals for predictions
        P = number of traits
    """

    def __init__(self, Y, F=None, A=None, Fstar=None):
        """
        Args:
            Y:        phenotype matrix [N, P]
            F:        list of sample fixed effect designs.
                      Each term must have first dimension N
            A:        list of trait fixed effect design.
                      Each term must have second dimension P
            Fstar:    list out-of-sample fixed effect design.
                      Each term must have first dimension No
        """
        Cached.__init__(self)
        Y = assert_make_float_array(Y, 'Y')
        if F is not None:
            try:
                assert_type_or_list_type(F, np.ndarray, 'F')
            except TypeError as e:
                raise TypeError(e.message + ' Parameter F might also be set'
                                ' to None.')

        if A is not None:
            try:
                assert_type_or_list_type(A, np.ndarray, 'A')
            except TypeError as e:
                raise TypeError(e.message + ' Parameter A might also be set'
                                ' to None.')

        assert Fstar is None, 'This constructor still does not support Fstar.'

        self.Y = Y
        self.setDesigns(F, A)
        self.Fstar = Fstar
        self.setFIinv(None)
        self._set_relay(None)

    #########################################
    # Properties
    #########################################
    @property
    def n_terms(self):
        return self._n_terms

    @property
    def n_covs(self):
        return self._n_covs

    @property
    def Y(self):
        return self._Y

    @property
    @cached('pheno')
    def y(self):
        r = vec(self.Y)
        if self._miss:
            r = r[self._veIok]
        return r

    @property
    def F(self):
        return self._F

    @property
    def A(self):
        return self._A

    @property
    def B(self):
        B = []
        istart = 0
        for ti in range(self.n_terms):
            iend = istart + self.F[ti].shape[1] * self.A[ti].shape[0]
            B.append(sp.reshape(self.b[istart:iend], (self.F[ti].shape[1],
                     self.A[ti].shape[0]), order='F'))
            istart = iend
        return B

    @property
    def B_ste(self):
        print('TODO: implement me')

    @property
    def b_ste(self):
        if self.getFIinv() is None:
            R = None
        else:
            R = sp.sqrt(self.getFIinv().diagonal())[:, sp.newaxis]
        return R

    @property
    def n_data_points(self):
        return (~self._miss).sum()

    @property
    def Fstar(self):
        print('TODO: assert stuff')
        return self._Fstar

    @property
    @cached('designs')
    def W(self):
        R = sp.zeros((self.Y.size, self.n_covs))
        istart = 0
        for ti in range(self.n_terms):
            iend = istart + self.F[ti].shape[1] * self.A[ti].shape[0]
            R[:, istart:iend] = sp.kron(self.A[ti].T, self.F[ti])
            istart += iend
        if self._miss:
            R = R[self._veIok, :]
        return R

    @property
    def use_to_predict(self):
        return self._use_to_predict

    #########################################
    # Utils function
    #########################################
    def Ft_dot(self, M):
        dim0 = 0
        R = sp.zeros((self._k, M.shape[1]))
        istart = 0
        for ti in range(self.n_terms):
            _dim = self.F[ti].shape[1]
            iend = istart + _dim
            R[istart:iend] = sp.dot(self.F[ti].T, M)
            istart = iend
        return R

    #########################################
    # Setters
    #########################################
    @Y.setter
    def Y(self, value):
        """ set phenotype """
        self._N = value.shape[0]
        self._P = value.shape[1]
        self._Y = value
        # missing data
        self._Iok = ~sp.isnan(value)
        self._veIok = vec(self._Iok)[:,0]
        self._miss = (~self._Iok).any()
        # notify and clear_cached
        self.clear_cache('pheno')
        self._notify()
        self._notify('pheno')

    @y.setter
    def y(self, value):
        print(("%s: y.setter not available in this class"%(self.__class__)))

    def setDesigns(self, F, A):
        """ set fixed effect designs """
        F = to_list(F)
        A = to_list(A)
        assert len(A) == len(F), 'MeanKronSum: A and F must have same length!'
        n_terms = len(F)
        n_covs = 0
        k = 0
        l = 0
        for ti in range(n_terms):
            assert F[ti].shape[0] == self._N, 'MeanKronSum: Dimension mismatch'
            assert A[ti].shape[1] == self._P, 'MeanKronSum: Dimension mismatch'
            n_covs += F[ti].shape[1] * A[ti].shape[0]
            k += F[ti].shape[1]
            l += A[ti].shape[0]
        self._n_terms = n_terms
        self._n_covs = n_covs
        self._k = k
        self._l = l
        self._F = F
        self._A = A
        self._b = sp.zeros((n_covs, 1))
        self.clear_cache('predict_in_sample', 'Yres', 'designs')
        self._notify('designs')
        self._notify()

    @Fstar.setter
    def Fstar(self, value):
        """ set fixed effect design for predictions """
        if value is None:
            self._use_to_predict = False
        else:
            assert value.shape[1] == self._K, 'Dimension mismatch'
            self._use_to_predict = True
        self._Fstar = value
        self.clear_cache('predict')

    @use_to_predict.setter
    def use_to_predict(self, value):
        assert not (self.Fstar is None and value is True), 'set Fstar!'
        self._use_to_predict = value

    #########################################
    # Predictions
    #########################################
    @cached
    def predict(self):
        r = self._predict_fun(self.Fstar)
        return r

    @cached
    def predict_in_sample(self):
        r = self._predict_fun(self.F)
        return r

    def _predict_fun_mv(self, M):
        assert len(M) == self.n_terms, 'MeanKronSum: Dimension mismatch'
        rv = sp.zeros((self._N, self._P))
        for ti in range(self.n_terms):
            rv += sp.dot(sp.dot(M[ti], self.B[ti]), self.A[ti])
        return rv

    def _predict_fun(self, M):
        assert len(M) == self.n_terms, 'MeanKronSum: Dimension mismatch'
        rv = sp.zeros((self._N, self._P))
        for ti in range(self.n_terms):
            rv += sp.dot(sp.dot(M[ti], self.B[ti]), self.A[ti])
        return rv

    @cached('Yres')
    def Yres(self):
        return self.Y - self.predict_in_sample()

    @cached('Yres')
    def yres(self):
        r = vec(self.Yres())
        if self._miss:
            r = r[~self._veIok]
        return r

if __name__ == '__main__':

    # define phenotype
    N = 1000
    P = 4
    Y = sp.randn(N, P)

    # define fixed effects
    F = []
    A = []
    F.append(sp.randn(N, 3))
    F.append(sp.randn(N, 2))
    A.append(sp.eye(P))
    A.append(sp.ones((1, P)))

    pdb.set_trace()

    mean = MeanKronSum(Y, F, A)



import sys

try:
    sys.path.insert(0,'./../../..')
    sys.path.insert(0,'./../../../../../mksum/mtSet')
    from mtSet.pycore.gp import gp2kronSum as gp2kronSumMtSet
    import mtSet.pycore.mean as MEAN
    mtSet_present = True
except:
    print("no mtSet found in path")
    mtSet_present = False

from limix.core.mean import mean
from limix.core.mean.mean_efficient import Mean
from limix.core.gp import gp2kronSum as gp2kronSum

from limix.core.covar import freeform 
import limix.core.optimize.optimize_bfgs as OPT
import limix.core.association.lmm_kronecker as lmm

import scipy as SP
import scipy.linalg as LA 
import time as TIME
import copy




if __name__ == "__main__":

    plot = "plot" in sys.argv

    # generate data
    h2 = 0.999
    N = 500; P = 2; S = 1000
    X = 1.*(SP.rand(N,S)<0.2)
    snps = 1.*(SP.rand(N,S)<0.2)
    snps_0 = 1.*(SP.rand(N,S)<0.2)
    beta = SP.randn(S,P)
    beta_snp = SP.randn(S,P)*10000
    
    #beta_snp[(SP.rand(beta_snp.shape[0],beta_snp.shape[1])<0.1)]=0.0

    Yg = SP.dot(X,beta);
    Yg += SP.dot(snps,beta_snp);
    Yg*=SP.sqrt(h2/Yg.var(0).mean())
    Yn = SP.randn(N,P); Yn*=SP.sqrt((1-h2))
    Y  = Yg+Yn; Y-=Y.mean(0); Y/=Y.std(0)
    XX = SP.dot(X,X.T)
    XX/= XX.diagonal().mean()
    Xr = 1.*(SP.rand(N,10)<0.2)
    Xr*= SP.sqrt(N/(Xr**2).sum())

    # define mean term
    mu = mean(Y)

    # add first fixed effect
    F = 1.*(SP.rand(N,3)<0.2); A = SP.eye(P)
    F[:,0]=1.0

    # add second fixed effect
    F2 = 1.*(SP.rand(N,4)<0.2); A2 = SP.ones((1,P))
    
    mu.addFixedEffect(F=F,A=A)
    mu.addFixedEffect(F=F2,A=A2)
    

    # define covariance matrices
    Cg = freeform(P)
    Cn = freeform(P)
    Cg1 = freeform(P)
    Cn1 = freeform(P)
    
    # compare with mtSet implementation
    params = {}
    params['Cg']   = SP.randn(int(0.5*P*(P+1)))
    params['Cn']   = SP.randn(int(0.5*P*(P+1)))


    print("check gradient with gp2kronSum")
    gp = gp2kronSum(mu,Cg,Cn,XX)
    gp.setParams(params)

    if "ML" in sys.argv:
        gp.set_reml(False)
    if 1:#optimizwe
        print("test optimization")
        start = TIME.time()
        conv,info = OPT.opt_hyper(gp,params,factr=1e3)
        print(('Reml GP:', TIME.time()-start))
        

        print(conv)

    #association
    assoc = lmm.LmmKronecker(gp=gp)
    if 1:
        #signal snps
        pv,LL_snps,LL_snps_0= assoc.test_snps(snps, identity_trick = True)
        #null snps
        pv0,n_LL_snps0,LL_snps0_0 = assoc.test_snps(snps_0, identity_trick = True)
    
    if plot:
        import pylab as pl
        pl.ion()
        #pl.figure(); pl.hist(pv,50)
        #pl.figure(); pl.hist(pv0,50)
        i_pv = pv.argsort()
        i_pv0 = pv0.argsort()
        pl.figure()
        pl.plot(-SP.log(pv[i_pv]),-SP.log(pv0[i_pv0]),'.')
        pl.plot([0,8],[0,8])

        pv0_,n_LL_snps0_,LL_snps0_0_ = assoc.test_snps(snps_0, identity_trick = True)
        pl.figure(); pl.plot(-SP.log(pv0_),-SP.log(pv0),'.')
        pl.figure(); pl.plot(-SP.log(pv0_),-SP.log(pv0),'.')

    if 1:
        lml0 = assoc._gp.LMLdebug()
        lml1 = assoc._gp.LML()  
        gp_ = assoc._gp
        var_expl1,beta1 = assoc._gp.mean.var_explained()
        

        
        var_expl1,beta1 = assoc._gp.mean.var_explained()
        mu_ = Mean(Y)
        mu_.addFixedEffect(F=F,A=A)
        mu_.addFixedEffect(F=F2,A=A2)
        gp_.setMean(mu_)
        lml2 = gp_.LML()
        var_expl2,beta_hat, beta_hat_any = assoc._gp.mean.var_explained()

        mu__ = mean(Y,identity_trick=True)
        mu__.addFixedEffect(F=F,A=A)
        mu__.addFixedEffect(F=F2,A=A2)
        gp_.setMean(mu__)
        var_expl3,beta3 = assoc._gp.mean.var_explained()
        print(lml0)#complete garbage
        print(lml1)
        print(lml2)

    if 0:
        #forward selection step
        assoc.addFixedEffect(F=snps[:,i_pv[0]:(i_pv[0]+1)],A=None)
        
        if 1:#optimize
            print("test optimization")
            start = TIME.time()
            conv,info = OPT.opt_hyper(assoc._gp,params,factr=1e3)
            print(('Reml GP:', TIME.time()-start))
        

        print(conv)
        pv_forw,LL_snps_forw,LL_snps_0_forw = assoc.test_snps(snps,identity_trick = True)
        if plot:
            pl.figure()
            pl.plot(-SP.log(pv_forw),-SP.log(pv),'.')
            pl.plot([0,8],[0,8])





# import sys
# import scipy as SP
# import scipy.linalg as LA
# sys.path.insert(0,'/Users/casale/Documents/limix/limix')
# from limix.core.covar import freeform
# from limix.core.gp.gp3kronSumApprox import gp3kronSumApprox
# import limix.core.optimize.optimize_bfgs as OPT
# sys.path.append('./../../../build/release.darwin/interfaces/python/limix/modules')
# import varianceDecomposition as VAR
#
# import h5py
# import pylab as PL
# PL.ion()
#
# def genPheno(G=None,X=None,var_g=None,var_x=None,P=None):
#     var_n = 1-var_g-var_x
#     Yg = SP.dot(G,SP.randn(G.shape[1],P))
#     Yx = SP.dot(X,SP.randn(X.shape[1],P))
#     Yn = SP.randn(X.shape[0],P)
#     Yg*= SP.sqrt(var_g/Yg.var(0).mean())
#     Yx*= SP.sqrt(var_x/Yx.var(0).mean())
#     Yn*= SP.sqrt(var_n/Yn.var(0).mean())
#     RV = Yg+Yx+Yn
#     RV-= RV.mean(0); RV/= RV.std(0)
#     return RV
#
# if __name__=='__main__':
#
#     P = 4
#
#     seed = int(sys.argv[1])
#     SP.random.seed(seed)
#
#     # import data
#     fname = 'data/arab107_preprocessed.hdf5'
#     f = h5py.File(fname,'r')
#     X = f['genotype'][:]
#     X-= X.mean(0); X/=X.std(0)
#     G = X[:,0:200]
#
#     Y = genPheno(G=G,X=X,var_g=0.10,var_x=0.40,P=P)
#     XX = SP.dot(X,X.T)
#     XX/= XX.diagonal().mean()
#     XX+= 1e-4*SP.eye(XX.shape[0])
#     GG = SP.dot(G,G.T)
#     GG/= GG.diagonal().mean()
#     GG+= 1e-4*SP.eye(GG.shape[0])
#
#     Cr = freeform(P)
#     Cg = freeform(P)
#     Cn = freeform(P)
#     gp = gp3kronSumApprox(Y=Y,Cr=Cr,Cg=Cg,Cn=Cn,XX=XX,GG=GG)
#     n_rips = 10
#     for rip in range(n_rips):
#
#         Cr.setRandomParams()
#         Cg.setRandomParams()
#         Cn.setRandomParams()
#         params = gp.getParams()
#         gp.setParams(params)
#
#         conv,info = OPT.opt_hyper(gp,params,factr=1e3)
#         print(conv)
#
#         print('Cr')
#         print((Cr.K()))
#         print('Cg')
#         print((Cg.K()))
#         print('Cn')
#         print((Cn.K()))
#
#     ipdb.set_trace()
#
#     gp.setBound(0)
#     conv,info = OPT.opt_hyper(gp,params,factr=1e3)
#     print(conv)
#
#     print('Cr')
#     print((Cr.K()))
#     print('Cg')
#     print((Cg.K()))
#     print('Cn')
#     print((Cn.K()))
#
#     gp.setBound(1)
#     conv,info = OPT.opt_hyper(gp,params,factr=1e3)
#     print(conv)
#
#     print('Cr')
#     print((Cr.K()))
#     print('Cg')
#     print((Cg.K()))
#     print('Cn')
#     print((Cn.K()))
#
#     ipdb.set_trace()
#




import h5py
import pandas
from optparse import OptionParser
import scipy as sp
import os
import pdb
import subprocess, sys, os
import limix.deprecated.io.plink as PLINK


class LIMIX_converter(object):
    '''
    A class to help with file conversion in LIMIX
    '''
    __slots__=["options","args","data", "result","infostring","hdf"]

    def __init__(self,infostring=None):
        '''
        nothing to initialize
        '''
        self.options=None
        self.args=None
        self.data=None
        self.result={}
        self.infostring=infostring
        if self.infostring is not None:
            self.result["infostring"]=self.infostring
        pass

    def parse_args(self):
        usage = "usage: %prog [options]"
        parser = OptionParser(usage=usage)
        parser.add_option("-O","--outfile", action="store", dest='outfile', type=str, help='The output hdf5 file wiht the resulting data', default="example_out")
        parser.add_option("-P","--plink",action="store", dest='plink', help="Read genotype from binary plink file (filename)", default=None)
        parser.add_option("-G","--g012",action="store", dest='g012', help="Read genotype from 012 file generated using vcftools (filename)", default=None)
        parser.add_option("-d","--g012_dosage",action="store", dest='g012_dosage', help="Read genotype dosage file generated using vcftools (filename)", default=None)
        parser.add_option("-c","--chrom",action="store", dest='chrom', help="Select specific chromosome for conversion (default: all)", default=None)
        parser.add_option("-s","--start",action="store", dest='start', help="Specify start position for conversion (default: all)", default=None)
        parser.add_option("-e","--end",action="store", dest='end', help="Specify start position for conversion (default: all)", default=None)
        parser.add_option("-C","--csv",action="store", dest='csv', help="Phenotype csv file", default=None)
        parser.add_option("-T","--csv_transpose",action="store_true", dest='csv_transpose', help="transpose phenotpye csv file (standard is rows: samples)", default=False)
        parser.add_option("-D","--csv_sep",action="store", dest='csv_sep', help="CSV delimiter (standard: auto)", default=None)
        parser.add_option("-S","--csv_num_samples",action="store", dest='csv_num_samples', help="Number of samples to parse (all)", default=None)
        parser.add_option("-Q","--csv_num_phenotypes",action="store", dest='csv_num_phenotypes', help="Number of phenotypes to parse (all)", default=None)

        (self.options, self.args) = parser.parse_args()
        self.result["options"]=str(self.options)
        return (self.options,self.args)

    def run(self):
        """
        Run the conversion job
        """
        #open hdfstore
        self.hdf = h5py.File(self.options.outfile)

        #read genotype plink file?
        if self.options.csv is not None:
            self.convert_phenotype_csv(self.hdf,self.options.csv,self.options.csv_sep,self.options.csv_transpose,num_samples=self.options.csv_num_samples,num_phenotypes=self.options.csv_num_phenotypes)
        if self.options.plink is not None:
            self.convert_plink(self.hdf,self.options.plink,self.options.chrom,self.options.start,self.options.end)
        if self.options.g012 is not None:
            self.convert_g012(self.hdf,self.options.g012,self.options.chrom,self.options.start,self.options.end)
        if self.options.g012_dosage is not None:
            self.convert_g012_dosage(self.hdf,self.options.g012_dosage,self.options.chrom,self.options.start,self.options.end)
        pass

    def convert_phenotype_csv(self,hdf,csv_file,sep=None,transpose=False,num_samples=None,num_phenotypes=None,*args,**kw_args):
        """
        convert phenotype csv file to LIMIX hdf5

        Arguments:
            hdf: handle for hdf5 file (target)
            csv_file: filename of csv file with phenoyptes
            transpose: standard is rows - individuals unless transpose = True
        """
        C = pandas.io.parsers.read_csv(csv_file,sep=sep,index_col=[0],*args,**kw_args)
        if transpose:
            C = C.T
        matrix = C.values
        #TODO: check if sample_IDs are present, and if they may be 2 columns
        sample_IDs = C.index.values.astype(str)
        #TODO: check if pheno_IDs are present
        pheno_IDs  = C.columns.values.astype(str)
        if num_samples is not None:
            matrix = matrix[0:1+int(num_samples)]
            sample_IDs =  sample_IDs[0:1+int(num_samples)]
        if num_phenotypes is not None:
            matrix = values[:,0:1+int(num_phenotypes)]
            pheno_IDs = pheno_IDs[0:1+int(num_phenotypes)]
        #store
        if 'phenotype' in list(hdf.keys()):
            del(hdf['phenotype'])
        phenotype = hdf.create_group('phenotype')
        col_header = phenotype.create_group('col_header')
        row_header = phenotype.create_group('row_header')
        #write phenotype
        phenotype.create_dataset(name='matrix',data=matrix)
        #write row header
        row_header.create_dataset(name='sample_ID',data=sample_IDs)
        #write col header
        col_header.create_dataset(name='phenotype_ID',data=pheno_IDs)
        pass


    def convert_g012(self,hdf,g012_file,chrom,start,end):
        """convert g012 file to LIMIX hdf5
        hdf: handle for hdf5 file (target)
        g012_file: filename of g012 file
        chrom: select chromosome for conversion
        start: select start position for conversion
        end:  select end position for conversion
        """
        if ((start is not None) or (end is not None) or (chrom is not None)):
            print("cannot handle start/stop/chrom boundaries for g012 file")
            return
        #store
        if 'genotype' in list(hdf.keys()):
            del(hdf['genotype'])
        genotype = hdf.create_group('genotype')
        col_header = genotype.create_group('col_header')
        row_header = genotype.create_group('row_header')
        #load position and meta information
        indv_file = g012_file + '.indv'
        pos_file  = g012_file + '.pos'
        sample_ID = sp.loadtxt(indv_file,dtype='str')
        pos  = sp.loadtxt(pos_file,dtype='str')
        chrom = pos[:,0]
        pos   = sp.array(pos[:,1],dtype='int')

        row_header.create_dataset(name='sample_ID',data=sample_ID)
        col_header.create_dataset(name='chrom',data=chrom)
        col_header.create_dataset(name='pos',data=pos)
        M = sp.loadtxt(g012_file,dtype='uint8')
        snps = M[:,1::]
        genotype.create_dataset(name='matrix',data=snps,chunks=(snps.shape[0],min(10000,snps.shape[1])),compression='gzip')
        pass

    def convert_g012_dosage(self,hdf,g012_file,chrom,start,end):
        """convert g012 file to LIMIX hdf5
        hdf: handle for hdf5 file (target)
        g012_file: filename of g012 file
        chrom: select chromosome for conversion
        start: select start position for conversion
        end:  select end position for conversion
        """
        if ((start is not None) or (end is not None) or (chrom is not None)):
            print("cannot handle start/stop/chrom boundaries for g012 file")
            return
        #store
        if 'genotype' in list(hdf.keys()):
            del(hdf['genotype'])
        genotype = hdf.create_group('genotype')
        col_header = genotype.create_group('col_header')
        row_header = genotype.create_group('row_header')
        #load position and meta information
        indv_file = g012_file + '.indv'
        pos_file  = g012_file + '.pos'
        sample_ID = sp.loadtxt(indv_file,dtype='str')
        pos  = sp.loadtxt(pos_file,dtype='str')
        chrom = pos[:,0]
        pos   = sp.array(pos[:,1],dtype='int')
        # load marker info
        ids_file = g012_file + '.ids'
        ids = sp.loadtxt(ids_file,dtype='str')
        gdid = ids[:,2]
        ref = ids[:,3]
        alt = ids[:,4]
        impute = ids[:,5]
        row_header.create_dataset(name='sample_ID',data=sample_ID)
        col_header.create_dataset(name='chrom',data=chrom)
        col_header.create_dataset(name='pos',data=pos)
        col_header.create_dataset(name='gdid',data=gdid)
        col_header.create_dataset(name='ref',data=ref)
        col_header.create_dataset(name='alt',data=alt)
        col_header.create_dataset(name='IMP2',data=impute)
        M = sp.loadtxt(g012_file,dtype='float')
        snps = M[:,:-1]
        genotype.create_dataset(name='matrix',data=snps,chunks=(snps.shape[0],min(10000,snps.shape[1])),compression='gzip')
        pass

    def convert_plink(self,hdf,bed_file,chrom,start,end):
        """convert plink file to LIMIX hdf5
        hdf: handle for hdf5 file (target)
        bed_file: filename of bed file
        chrom: select chromosome for conversion
        start: select start position for conversion
        end:  select end position for conversion
        """
        #TODO: check that this is a valid plink file
        #start_pos?
        startpos = None
        endpos   = None
        if (start is not None) or (end is not None):
            if chrom is None:
                raise Exception('start or end pos specification requires a target chromosome')
        if start is not None:
            startpos = [chrom,0,start]
        if end is not None:
            endpos = [chrom,0,end]

        #read
        data=PLINK.readBED(bed_file,startpos=startpos,endpos=endpos)
        #store
        if 'genotype' in list(hdf.keys()):
            del(hdf['genotype'])
        genotype = hdf.create_group('genotype')
        col_header = genotype.create_group('col_header')
        row_header = genotype.create_group('row_header')
        #write genotype
        genotype.create_dataset(name='matrix',data=data['snps'],chunks=(data['snps'].shape[0],min(10000,data['snps'].shape[1])))
        #write row header
        row_header.create_dataset(name='sample_ID',data=data['iid'][:,0])
        #write col header
        col_header.create_dataset(name='chrom',data=data['pos'][:,0])
        col_header.create_dataset(name='pos',data=data['pos'][:,2])
        col_header.create_dataset(name='alleles',data=data['alleles'])
        pass



import numpy as np
import scipy as sp
import scipy.linalg as LA
from .covar_base import Covariance
from hcache import cached
import pdb

import logging as LG

class DiagonalCov(Covariance):
    """
    Diagonal covairance matrix
    """
    def __init__(self, dim, jitter=1e-4):
        """
        Args:
            dim:        dimension of the free-form covariance
            jitter:     extent of diagonal offset which is added for numerical stability
                        (default value: 1e-4)
        """
        Covariance.__init__(self, dim)
        self._K_act = True
        self._calcNumberParams()
        self.dim = dim
        self.params = sp.zeros(self.n_params)
        self.set_jitter(jitter)

    #####################
    # Properties
    #####################
    @property
    def variance(self):
        return sp.exp(self.params) 

    @property
    def variance_ste(self):
        #TODO
        pass

    @property
    def X(self):
        return sp.sqrt(self.K())

    #####################
    # Activation handling
    #####################
    @property
    def act_K(self):
        return self._K_act

    @act_K.setter
    def act_K(self, act):
        self._K_act = bool(act)
        self._notify()

    #####################
    # Params handling
    #####################
    def setParams(self, params):
        if not self._K_act and len(params) > 0:
            raise ValueError("Trying to set a parameter via setParams that "
                             "is not active.")
        if self._K_act:
            self.params[:] = params
            self.clear_all()

    def getParams(self):
        if not self._K_act:
            return np.array([])
        return self.params

    def getNumberParams(self):
        return int(self._K_act) * self.n_params

    def _calcNumberParams(self):
        self.n_params = self.dim

    def set_jitter(self,value):
        self.jitter = value

    def setCovariance(self,cov):
        """ set hyperparameters from given covariance """
        self.setParams(sp.log(sp.diagonal(cov)))

    #####################
    # Cached
    #####################
    @cached('covar_base')
    def K(self):
        return sp.diag(self.variance)

    @cached('covar_base')
    def K_grad_i(self,i):
        if not self._K_act:
            raise ValueError("Trying to retrieve the gradient over a "
                             "parameter that is inactive.")
        RV = sp.zeros((self.dim, self.dim)) 
        RV[i, i] = self.variance[i]
        return RV

    @cached
    def K_hess_i_j(self, i, j):
        if not self._K_act:
            raise ValueError("Trying to retrieve the gradient over a "
                             "parameter that is inactive.")
        RV = sp.zeros((self.dim, self.dim)) 
        if i==j:    RV[i, i] = self.variance[i]
        return RV

    def K_ste(self):
        #TODO
        #if self.getFIinv() is None:
        #    R = None
        #else:
        #    R = sp.zeros((self.dim, self.dim))
        #    R[sp.tril_indices(self.dim)] = sp.sqrt(self.getFIinv().diagonal())
        #    # symmetrize
        #    R = R + R.T - sp.diag(R.diagonal())
        #return R
        pass

    ####################
    # Interpretable Params
    ####################
    def getInterParams(self):
        return self.variance

    # DERIVARIVE WITH RESPECT TO COVARIANCES
    def K_grad_interParam_i(self, i):
        R = sp.zeros((self.dim,self.dim))
        R[i, i] = 1
        return R




import sys

import scipy as SP
import scipy.linalg
import scipy.stats
import limix
import limix.utils.preprocess as preprocess
import pdb
import time
import copy

class VarianceDecomposition:
    """
    Variance decomposition module in LIMIX
    This class mainly takes care of initialization and eases the interpretation of complex variance decompositions

    #TODO: change documentation. Python modules explain what the functions do but a list of function is not needed, I think?
    #A 3 line example how to use it would be good to have here... perhaps.
    Methods:
        __init__(self,Y):            Constructor
        addSingleTraitTerm:          Add Single Trait Term
        addMultiTraitTerm:           Add Multi Trait Term (form of trait-to-trait covariance matrix need to be set as well)
        addFixedTerm:                Add Fixed Effect Term
        fit:                         Fit phenos and returns the minimum with some info
        setScales:                   Set the vector of covar parameterss
        getScales:                   Return the vector of the covar parameters
        getFixed:                    Return the vector of the dataTerm parameters
        getEmpTraitCovar:            Return the empirical trait covariance
        getEstTraitCovar:            Return the estimated trait-to-trait covariance matrix for term i
        estimateHeritabilities:      Given K, it fits the model with 1 global fixed effects and covariance matrix hg[p]*K+hn[p] for each trait p and returns the vectors hg and hn
    """
    
    
    def __init__(self,Y,standardize=False):
        """
        Args:
            Y:              phenotype matrix [N, P]
            standardize:    if True, impute missing phenotype values by mean value,
                            zero-mean and unit-variance phenotype (Boolean, default False)
        """
        
        #create column of 1 for fixed if nothing providede
        self.N       = Y.shape[0]
        self.P       = Y.shape[1]
        self.Nt      = self.N*self.P
        self.Iok     = ~(SP.isnan(Y).any(axis=1))

        #outsourced to handle missing values:
        if standardize:
            Y=preprocess.standardize(Y)

        self.Y       = Y
        self.vd      = limix.CVarianceDecomposition(Y)
        self.n_terms = 0
        self.gp      = None
        self.init    = False
        self.fast    = False
        self.noisPos = None
        self.optimum = None

        # for multi trait models 
        self.covar_type = []
        self.diag       = []
        self.offset     = []
        
        self.cache = {}
        self.cache['Sigma']   = None
        self.cache['Hessian'] = None
        self.cache['Lparams'] = None
        self.cache['paramsST']= None

        pass
    
    
    def setY(self,Y,standardize=False):
        """
        Set phenotype matrix
        
        Args:
            Y:              phenotype matrix [N, P]
            standardize:	if True, phenotype is standardized (zero mean, unit variance)
        """
        assert Y.shape[0]==self.N, 'CVarianceDecomposition:: Incompatible shape'
        assert Y.shape[1]==self.P, 'CVarianceDecomposition:: Incompatible shape'

        
        if standardize:
            Y=preprocess.standardize(Y)

        #check that missing values match the current structure
        assert (~(SP.isnan(Y).any(axis=1))==self.Iok).all(), 'CVarianceDecomposition:: pattern of missing values needs to match Y given at initialization'

        self.Y = Y
        self.vd.setPheno(Y)

        self.optimum = None
    
        self.cache['Sigma']   = None
        self.cache['Hessian'] = None
        self.cache['Lparams'] = None
        self.cache['paramsST']= None

    def addRandomEffect(self,K=None,covar_type='freeform',is_noise=False,normalize=True,Ks=None,offset=1e-4,rank=1,covar_K0=None):
        """
        Add random effect Term
        depending on self.P=1 or >1 add single trait or multi trait random effect term
        """
        if self.P==1:	self.addSingleTraitTerm(K=K,is_noise=is_noise,normalize=normalize,Ks=Ks)
        else:			self.addMultiTraitTerm(K=K,covar_type=covar_type,is_noise=is_noise,normalize=normalize,Ks=Ks,offset=offset,rank=rank,covar_K0=covar_K0)

    
    def addSingleTraitTerm(self,K=None,is_noise=False,normalize=True,Ks=None):
        """
        add random effects term for single trait models (no trait-trait covariance matrix)
        
        Args:
            K:          NxN sample covariance matrix
            is_noise:	bool labeling the noise term (noise term has K=eye)
            normalize:	if True, K and Ks are scales such that K.diagonal().mean()==1
            Ks:			NxN test cross covariance for predictions
        """
        
        assert self.P == 1, 'Incompatible number of traits'
        
        assert K!=None or is_noise, 'Specify covariance structure'
        
        if is_noise:
            assert self.noisPos==None, 'noise term already exists'
            K = SP.eye(self.Nt)
            self.noisPos = self.n_terms
        else:
            assert K.shape[0]==self.Nt, 'Incompatible shape'
            assert K.shape[1]==self.Nt, 'Incompatible shape'
    
        if Ks!=None:
            assert Ks.shape[0]==self.N, 'Incompatible shape'

        if normalize:
            Norm = 1/K.diagonal().mean()
            K *= Norm
            if Ks!=None: Ks *= Norm

        self.vd.addTerm(limix.CSingleTraitTerm(K))
        if Ks!=None: self.setKstar(self.n_terms,Ks)
        self.n_terms+=1
    
        self.gp         = None
        self.init       = False
        self.fast       = False
        self.optimum    = None

        self.cache['Sigma']   = None
        self.cache['Hessian'] = None
        self.cache['Lparams'] = None
        self.cache['paramsST']= None

    
    #TODO: rename offset. Perhaps we can call it regularization or whiteNoise or something. 
    def addMultiTraitTerm(self,K=None,covar_type='freeform',is_noise=False,normalize=True,Ks=None,offset=1e-4,rank=1,covar_K0=None):
        """
        add multi trait random effects term.
        The inter-trait covariance is parametrized by covar_type, where parameters are optimized.
        
        Args:
            K:      Individual-individual (Intra-Trait) Covariance Matrix [N, N]
                    (K is normalised in the C++ code such that K.trace()=N)
            covar_type: type of covaraince to use. Default 'freeform'. possible values are 
                            'freeform': free form optimization, 
                            'fixed': use a fixed matrix specified in covar_K0,
                            'diag': optimize a diagonal matrix, 
                            'lowrank': optimize a low rank matrix. The rank of the lowrank part is specified in the variable rank,
                            'lowrank_id': optimize a low rank matrix plus the weight of a constant diagonal matrix. The rank of the lowrank part is specified in the variable rank, 
                            'lowrank_diag': optimize a low rank matrix plus a free diagonal matrix. The rank of the lowrank part is specified in the variable rank, 
                            'block': optimize the weight of a constant P x P block matrix of ones,
                            'block_id': optimize the weight of a constant P x P block matrix of ones plus the weight of a constant diagonal matrix,
                            'block_diag': optimize the weight of a constant P x P block matrix of ones plus a free diagonal matrix,                            
            is_noise:   Boolean indicator specifying if the matrix is homoscedastic noise (weighted identity covariance) (default False)
            normalize:  Boolean indicator specifying if K is normalized such that K.trace()=N.
            Ks:			NxNtest cross covariance for predictions
            offset:		diagonal contribution added to trait-to-trait covariance matrices for regularization
            rank:       rank of a possible lowrank component (default 1)
            covar_K0:   PxP matrix for the (predefined) trait-to-trait covariance matrix if fixed type is used
        """
        assert self.P > 1, 'CVarianceDecomposition:: Incompatible number of traits'
        assert K!=None or is_noise, 'CVarianceDecomposition:: Specify covariance structure'
        assert offset>=0, 'CVarianceDecomposition:: offset must be >=0'

        #TODO: check that covar_K0 is correct if fixed typeCF is used..

        if is_noise:
            assert self.noisPos==None, 'CVarianceDecomposition:: noise term already exists'
            K = SP.eye(self.N)
            self.noisPos = self.n_terms
        else:
            assert K.shape[0]==self.N, 'CVarianceDecomposition:: Incompatible shape'
            assert K.shape[1]==self.N, 'CVarianceDecomposition:: Incompatible shape'

        if Ks!=None:
            assert Ks.shape[0]==self.N, 'CVarianceDecomposition:: Incompatible shape'

        if normalize:
            Norm = 1/K.diagonal().mean()
            K *= Norm
            if Ks!=None: Ks *= Norm

        cov = limix.CSumCF()
        if covar_type=='freeform':
            cov.addCovariance(limix.CFreeFormCF(self.P))
            L = SP.eye(self.P)
            diag = SP.concatenate([L[i,:(i+1)] for i in range(self.P)])
        elif covar_type=='fixed':
            cov.addCovariance(limix.CFixedCF(covar_K0))
            diag = SP.zeros(1)
        elif covar_type=='diag':
            cov.addCovariance(limix.CDiagonalCF(self.P))
            diag = SP.ones(self.P)
        elif covar_type=='lowrank':
            cov.addCovariance(limix.CLowRankCF(self.P,rank))
            diag = SP.zeros(self.P*rank)
        elif covar_type=='lowrank_id':
            cov.addCovariance(limix.CLowRankCF(self.P,rank))
            cov.addCovariance(limix.CFixedCF(SP.eye(self.P)))
            diag = SP.concatenate([SP.zeros(self.P*rank),SP.ones(1)])
        elif covar_type=='lowrank_diag':
            cov.addCovariance(limix.CLowRankCF(self.P,rank))
            cov.addCovariance(limix.CDiagonalCF(self.P))
            diag = SP.concatenate([SP.zeros(self.P*rank),SP.ones(self.P)])
        elif covar_type=='block':
            cov.addCovariance(limix.CFixedCF(SP.ones((self.P,self.P))))
            diag = SP.zeros(1)
        elif covar_type=='block_id':
            cov.addCovariance(limix.CFixedCF(SP.ones((self.P,self.P))))
            cov.addCovariance(limix.CFixedCF(SP.eye(self.P)))
            diag = SP.concatenate([SP.zeros(1),SP.ones(1)])
        elif covar_type=='block_diag':
            cov.addCovariance(limix.CFixedCF(SP.ones((self.P,self.P))))
            cov.addCovariance(limix.CDiagonalCF(self.P))
            diag = SP.concatenate([SP.zeros(1),SP.ones(self.P)])
        else:
            assert True==False, 'CVarianceDecomposition:: covar_type not valid'

        if offset>0:
            _cov = limix.CFixedCF(SP.eye(self.P))
            _cov.setParams(SP.array([SP.sqrt(offset)]))
            _cov.setParamMask(SP.zeros(1))
            cov.addCovariance(_cov)
        self.offset.append(offset)

        self.covar_type.append(covar_type)
        self.diag.append(diag)

        self.vd.addTerm(cov,K)
        if Ks!=None: self.setKstar(self.n_terms,Ks)
        self.n_terms+=1

        self.gp      = None
        self.init    = False
        self.fast    = False
        self.optimum = None

        self.cache['Sigma']   = None
        self.cache['Hessian'] = None
        self.cache['Lparams'] = None
        self.cache['paramsST']= None
    
    
    def addFixedEffect(self,F=None,A=None):
        """
        add fixed effect to the model

        Args:
            F: fixed effect matrix [N,1]
            A: design matrix [K,P] (e.g. SP.ones((1,P)) common effect; SP.eye(P) any effect)
        """
        if A==None:
            A = SP.eye(self.P)
        if F==None:
            F = SP.ones((self.N,1))
        
        assert A.shape[1]==self.P, 'Incompatible shape'
        assert F.shape[0]==self.N, 'Incompatible shape'
       
        if F.shape[1]>1:
            for m in range(F.shape[1]):
                self.vd.addFixedEffTerm(A,F[:,m:m+1])
        else:
            self.vd.addFixedEffTerm(A,F)

        #TODO: what is this gp object doing, is this initialization correct?
        self.gp      = None
        self.init    = False
        self.fast    = False
        self.optimum = None

        self.cache['Sigma']   = None
        self.cache['Hessian'] = None
        self.cache['Lparams'] = None
        self.cache['paramsST']= None

    #TODO: automatically select the fast inference method if applicable. 
    #We can create a method "isFast()" to check whether the model is compatible with the fast version but really this can be automated.
    def initGP(self,fast=False):
        """
        Initialize GP objetct
        
        Args:
            fast:   if fast==True initialize gpkronSum gp
        """
        if fast:
            assert self.n_terms==2, 'CVarianceDecomposition: for fast inference number of terms must be == 2'
            assert self.P>1,        'CVarianceDecomposition: for fast inference number of traits must be > 1'
            self.vd.initGPkronSum()
        else:
            self.vd.initGP()
        self.gp=self.vd.getGP()
        self.init=True
        self.fast=fast

    #TODO: move all internal function to the end and flag them as such
    def _getScalesDiag(self,termx=0):
        """
        Uses 2 term single trait model to get covar params for initialization
        
        Args:
            termx:      non-noise term terms that is used for initialization 
        """
        assert self.P>1, 'CVarianceDecomposition:: diagonal init_method allowed only for multi trait models' 
        assert self.noisPos!=None, 'CVarianceDecomposition:: noise term has to be set'
        assert termx<self.n_terms-1, 'CVarianceDecomposition:: termx>=n_terms-1'
        assert self.covar_type[self.noisPos] not in ['lowrank','block','fixed'], 'CVarianceDecimposition:: diagonal initializaiton not posible for such a parametrization'
        assert self.covar_type[termx] not in ['lowrank','block','fixed'], 'CVarianceDecimposition:: diagonal initializaiton not posible for such a parametrization'
        scales = []
        res = self.estimateHeritabilities(self.vd.getTerm(termx).getK())
        scaleg = SP.sqrt(res['varg'].mean())
        scalen = SP.sqrt(res['varn'].mean())
        for term_i in range(self.n_terms):
            if term_i==termx:
                _scales = scaleg*self.diag[term_i]
            elif term_i==self.noisPos:
                _scales = scalen*self.diag[term_i]
            else:
                _scales = 0.*self.diag[term_i]
            if self.offset[term_i]>0:
                _scales = SP.concatenate((_scales,SP.array([SP.sqrt(self.offset[term_i])])))
            scales.append(_scales)
        return SP.concatenate(scales)

    def _getScalesRand(self):
        """
        Return a vector of random scales
        """
        if self.P>1:
            scales = []
            for term_i in range(self.n_terms):
                _scales = SP.randn(self.diag[term_i].shape[0])
                if self.offset[term_i]>0:
                    _scales = SP.concatenate((_scales,SP.array([SP.sqrt(self.offset[term_i])])))
                scales.append(_scales)
            scales = SP.concatenate(scales)
        else:
            scales=SP.randn(self.vd.getNumberScales())

        return scales

    def _perturbation(self):
        """
        Returns Gaussian perturbation
        """
        if self.P>1:
            scales = []
            for term_i in range(self.n_terms):
                _scales = SP.randn(self.diag[term_i].shape[0])
                if self.offset[term_i]>0:
                    _scales  = SP.concatenate((_scales,SP.zeros(1)))
                scales.append(_scales)
            scales = SP.concatenate(scales)
        else:
            scales = SP.randn(self.vd.getNumberScales())
        return scales
 
    def trainGP(self,fast=False,scales0=None,fixed0=None,lambd=None):
        """
        Train the gp
       
        Args:
            fast:       if true and the gp has not been initialized, initializes a kronSum gp
            scales0:	initial variance components params
            fixed0:     initial fixed effect params
        """
        assert self.n_terms>0, 'CVarianceDecomposition:: No variance component terms'

        if not self.init:		self.initGP(fast=fast)

        # set lambda
        if lambd!=None:		self.gp.setLambda(lambd)

        # set scales0
        if scales0!=None:
            self.setScales(scales0)
        # init gp params
        self.vd.initGPparams()
        # set fixed0
        if fixed0!=None:
            params = self.gp.getParams()
            params['dataTerm'] = fixed0
            self.gp.setParams(params)

        # LIMIX CVARIANCEDECOMPOSITION TRAINING
        conv =self.vd.trainGP()
        
        self.cache['Sigma']   = None
        self.cache['Hessian'] = None
            
        return conv

    #TOOD: auto select of fast (see initGP)
    #Q: is initGP not an internal method? Is the user going to call this himself? 
    def findLocalOptimum(self,fast=False,scales0=None,fixed0=None,init_method=None,termx=0,n_times=10,perturb=True,pertSize=1e-3,verbose=True,lambd=None):
        """
        Train the model using the specified initialization strategy
        
        Args:
            fast:		    if true, fast gp is initialized
            scales0:        if not None init_method is set to manual
            fixed0:         initial fixed effects
            init_method:    initialization method \in {random,diagonal,manual} 
            termx:			term for diagonal diagonalisation
            n_times:        number of times the initialization
            perturb:        if true, the initial point is perturbed with gaussian noise
            perturbSize:    size of the perturbation
            verbose:        print if convergence is achieved
        """

        if init_method==None:
            if self.P==1:	init_method = 'random'
            else:           init_method = 'diagonal'

        if not self.init:		self.initGP(fast=fast)

        if scales0!=None and ~perturb: 	init_method = 'manual'
        
        if init_method=='diagonal':
            scales0 = self._getScalesDiag(termx=termx)
            
        if init_method=='diagonal' or init_method=='manual':
            if not perturb:		n_times = 1

        if fixed0==None:
            fixed0 = SP.zeros_like(self.gp.getParams()['dataTerm'])

        for i in range(n_times):
            if init_method=='random':
                scales1 = self._getScalesRand()
                fixed1  = pertSize*SP.randn(fixed0.shape[0],fixed0.shape[1])
            elif perturb:
                scales1 = scales0+pertSize*self._perturbation()
                fixed1  = fixed0+pertSize*SP.randn(fixed0.shape[0],fixed0.shape[1])
            else:
                scales1 = scales0
                fixed1  = fixed0
            conv = self.trainGP(scales0=scales1,fixed0=fixed1,lambd=lambd)
            if conv:    break
    
        if verbose:
            if conv==False:
                print('No local minimum found for the tested initialization points')
            else:
                print(('Local minimum found at iteration %d' % i))
 
        return conv

    #TODO: this function and the above seem redundant. 
    #Don't you want this function to wrap the above, i.e. write the above as internal function and have this as the autside interface?
    #having multiple restarts is a simple add on. You can pass arguments automatically using the **kw_args keyword, which avoids having to redefine options here that are used again in the
    #internal function
    def findLocalOptima(self,fast=False,verbose=True,n_times=10,lambd=None):
        """
        Train the model repeadly up to a number specified by the users with random restarts and
        return a list of all relative minima that have been found 
        
        Args:
            fast:       Boolean. if set to True initalize kronSumGP
            verbose:    Boolean. If set to True, verbose output is produced. (default True)
            n_times:    number of re-starts of the optimization. (default 10)
        """
        if not self.init:       self.initGP(fast)
        
        opt_list = []

        fixed0 = SP.zeros_like(self.gp.getParams()['dataTerm'])    

        # minimises n_times
        for i in range(n_times):
            
            scales1 = self._getScalesRand()
            fixed1  = 1e-1*SP.randn(fixed0.shape[0],fixed0.shape[1])
            conv = self.trainGP(fast=fast,scales0=scales1,fixed0=fixed1,lambd=lambd)

            if conv:
                # compare with previous minima
                temp=1
                for j in range(len(opt_list)):
                    if SP.allclose(abs(self.getScales()),abs(opt_list[j]['scales'])):
                        temp=0
                        opt_list[j]['counter']+=1
                        break
                if temp==1:
                    opt = {}
                    opt['counter'] = 1
                    opt['LML'] = self.getLML()
                    opt['scales'] = self.getScales()
                    opt_list.append(opt)
        
        
        # sort by LML
        LML = SP.array([opt_list[i]['LML'] for i in range(len(opt_list))])
        index   = LML.argsort()[::-1]
        out = []
        if verbose:
            print("\nLocal mimima\n")
            print("n_times\t\tLML")
            print("------------------------------------")
            for i in range(len(opt_list)):
                out.append(opt_list[index[i]])
                if verbose:
                    print(("%d\t\t%f" % (opt_list[index[i]]['counter'], opt_list[index[i]]['LML'])))
                print("")

        return out

    #TODO: Q: externally visible function?
    def getLML(self):
        """
        Return log marginal likelihood
        """
        assert self.init, 'GP not initialised'
        return self.vd.getLML()


    #TODO: Q: externally visible function?
    def getLMLgrad(self):
        """
        Return gradient of log-marginal likelihood
        """
        assert self.init, 'GP not initialised'
        return self.vd.getLMLgrad()

    #TODO: text not clear get random ....
    def setScales(self,scales=None,term_num=None):
        """
        get random initialization of variances based on the empirical trait variance

        Args:
            scales:     if scales==None: set them randomly, 
                        else: set scales to term_num (if term_num==None: set to all terms)
            term_num:   set scales to term_num
        """
        if scales==None:
            for term_i in range(self.n_terms):
                n_scales = self.vd.getTerm(term_i).getNumberScales()
                self.vd.getTerm(term_i).setScales(SP.array(SP.randn(n_scales)))
        elif term_num==None:
            assert scales.shape[0]==self.vd.getNumberScales(), 'incompatible shape'
            index = 0
            for term_i in range(self.n_terms):
                index1 = index+self.vd.getTerm(term_i).getNumberScales()
                self.vd.getTerm(term_i).setScales(scales[index:index1])
                index = index1
        else:
            assert scales.shape[0]==self.vd.getTerm(term_num).getNumberScales(), 'incompatible shape'
            self.vd.getTerm(term_num).setScales(scales)

    #TODO: needs more text
    def getScales(self,term_i=None):
        """
        Returns the Parameters
        
        Args:
            term_i:     index of the term we are interested in
                        if term_i==None returns the whole vector of parameters
        """
        if term_i==None:
            RV = self.vd.getScales()
        else:
            assert term_i<self.n_terms, 'Term index non valid'
            RV = self.vd.getScales(term_i)
        return RV


    #TODO: naming? We use addFixedEffect. and getFixed. 
    def getFixed(self):
        """
        Return dataTerm params
        """
        assert self.init, 'GP not initialised'
        return self.gp.getParams()['dataTerm']


    def getEstTraitCovar(self,term_i=None):
        """
        Returns explicitly the estimated trait covariance matrix

        Args:
            term_i:     index of the term we are interested in
        """
        assert self.P>1, 'Trait covars not defined for single trait analysis'
        
        if term_i==None:
            RV=SP.zeros((self.P,self.P))
            for term_i in range(self.n_terms): RV+=self.vd.getTerm(term_i).getTraitCovar().K()
        else:
            assert term_i<self.n_terms, 'Term index non valid'
            RV = self.vd.getTerm(term_i).getTraitCovar().K()
        return RV


    #TODO: naming? suggests that this calculates one correlation coefficient rather than a matrix.
    #Add references between getEstTraitCorrCoeff and getEstTraitCovar
    def getEstTraitCorrCoef(self,term_i=None):
        """
        Returns the estimated trait correlation matrix

        Args:
            term_i:     index of the term we are interested in
        """
        cov = self.getEstTraitCovar(term_i)
        stds=SP.sqrt(cov.diagonal())[:,SP.newaxis]
        RV = cov/stds/stds.T
        return RV
    

    #TODO: naming? How does this relate to getScales?
    def getVariances(self):
        """
        Returns the estimated variances as a n_terms x P matrix
        each row of the output represents a term and its P values represent the variance corresponding variance in each trait
        """
        if self.P>1:
            RV=SP.zeros((self.n_terms,self.P))
            for term_i in range(self.n_terms):
                RV[term_i,:]=self.vd.getTerm(term_i).getTraitCovar().K().diagonal()
        else:
            RV=self.getScales()**2
        return RV


    #TODO: naming? How does this relate to getScales and getVariances?
    def getVarComponents(self):
        """
        Returns the estimated variance components as a n_terms x P matrix
        each row of the output represents a term and its P values represent the variance component corresponding variance in each trait
        """
        RV = self.getVariances()
        RV /= RV.sum(0)
        return RV

    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def getHessian(self):
        """
        COMPUTES OF HESSIAN OF E(\theta) = - log L(\theta | X, y)
        """
        assert self.init,        'GP not initialised'
        assert self.fast==False, 'Not supported for fast implementation'
        
        if self.cache['Hessian']==None:
            ParamMask=self.gp.getParamMask()['covar']
            std=SP.zeros(ParamMask.sum())
            H=self.gp.LMLhess_covar()
            It= (ParamMask[:,0]==1)
            self.cache['Hessian']=H[It,:][:,It]
        
        return self.cache['Hessian']
    
    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def getLaplaceCovar(self):
        """
        USES LAPLACE APPROXIMATION TO CALCULATE THE COVARIANCE MATRIX OF THE OPTIMIZED PARAMETERS
        """
        assert self.init,        'GP not initialised'
        assert self.fast==False, 'Not supported for fast implementation'
        
        if self.cache['Sigma']==None:
            self.cache['Sigma'] = SP.linalg.inv(self.getHessian())
        return self.cache['Sigma']

    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def getFisher(self):
        """
        Return the fisher information matrix over the parameters
        TO CHECK: IT DOES NOT WORK PROPERLY
        """
        Ctot = self.vd.getGP().getCovar()
        Ki = SP.linalg.inv(Ctot.K())
        n_scales = self.vd.getNumberScales()
        out = SP.zeros((n_scales,n_scales))
        for m in range(n_scales):
            out[m,m] = 0.5 * SP.trace(SP.dot(Ki,SP.dot(Ctot.Kgrad_param(m),SP.dot(Ki,Ctot.Kgrad_param(m)))))
            for n in range(m):
                out[m,n]=0.5 * SP.trace(SP.dot(Ki,SP.dot(Ctot.Kgrad_param(m),SP.dot(Ki,Ctot.Kgrad_param(n)))))
                out[n,m]=out[m,n]
        return out
    
    #TODO: naming. not clear what the standard errors refer to. I want getVarianceParams, getVarainceParamsSTtd() or similar.
    #one could also consider having std calculation as argument in the getVarianceParams.
    def getStdErrors(self,term_i):
        """
        RETURNS THE STANDARD DEVIATIONS ON VARIANCES AND CORRELATIONS BY PROPRAGATING THE UNCERTAINTY ON LAMBDAS

        Args:
            term_i:     index of the term we are interested in
        """
        assert self.init,        'GP not initialised'
        assert self.fast==False, 'Not supported for fast implementation'
        
        if self.P==1:
            out = (2*self.getScales()[term_i])**2*self.getLaplaceCovar()[term_i,term_i]
        else:
            C = self.vd.getTerm(term_i).getTraitCovar()
            n_params = C.getNumberParams()
            par_index = 0
            for term in range(term_i-1):
                par_index += self.vd.getTerm(term_i).getNumberScales()
            Sigma1 = self.getLaplaceCovar()[par_index:(par_index+n_params),:][:,par_index:(par_index+n_params)]
            out = SP.zeros((self.P,self.P))
            for param_i in range(n_params):
                out += C.Kgrad_param(param_i)**2*Sigma1[param_i,param_i]
                for param_j in range(param_i):
                    out += 2*abs(C.Kgrad_param(param_i)*C.Kgrad_param(param_j))*Sigma1[param_i,param_j]
        out = SP.sqrt(out)
        return out
    
    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def getModelPosterior(self,min):
        """
        USES LAPLACE APPROXIMATION TO CALCULATE THE BAYESIAN MODEL POSTERIOR
        """
        Sigma = self.getLaplaceCovar(min)
        n_params = self.vd.getNumberScales()
        ModCompl = 0.5*n_params*SP.log(2*SP.pi)+0.5*SP.log(SP.linalg.det(Sigma))
        RV = min['LML']+ModCompl
        return RV


    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def getEmpTraitCovar(self):
        """
        Returns the empirical trait covariance matrix
        """
        if self.P==1:
            out=self.Y[self.Iok].var()
        else:
            out=SP.cov(self.Y[self.Iok].T)
        return out


    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def getEmpTraitCorrCoef(self):
        """
        Returns the empirical trait correlation matrix
        """
        cov = self.getEmpTraitCovar()
        stds=SP.sqrt(cov.diagonal())[:,SP.newaxis]
        RV = cov/stds/stds.T
        return RV
    
    
    #TODO: externally visible function? THe external interface needs to be as static as possible. We should minimize this.
    def estimateHeritabilities(self, K, verbose=False):
        """
        estimate variance components and fixed effects
        from a single trait model having only two terms
        """
        
        # Fit single trait model
        varg  = SP.zeros(self.P)
        varn  = SP.zeros(self.P)
        fixed = SP.zeros((1,self.P))
        
        for p in range(self.P):
            y = self.Y[:,p:p+1]
            lmm = limix.CLMM()
            lmm.setK(K)
            lmm.setSNPs(SP.ones((K.shape[0],1)))
            lmm.setPheno(y)
            lmm.setCovs(SP.zeros((K.shape[0],1)))
            lmm.setVarcompApprox0(-20, 20, 1000)
            lmm.process()
            delta = SP.exp(lmm.getLdelta0()[0,0])
            Vtot  = SP.exp(lmm.getLSigma()[0,0])
    
            varg[p] = Vtot
            varn[p] = delta*Vtot
            fixed[:,p] = lmm.getBetaSNP()
    
            if verbose: print(p)
    
        sth = {}
        sth['varg']  = varg
        sth['varn']  = varn
        sth['fixed'] = fixed

        return sth

    """
    CODE FOR PREDICTIONS
    """
    
    #TODO: naming. not nice... needs also quite a few more checks, I think.
    def setKstar(self,term_i,Ks):
        """
        Set the kernel for predictions

        Args:
            term_i:     index of the term we are interested in
            Ks:         (TODO: is this the covariance between train and test or the covariance between test points?)
        """
        assert Ks.shape[0]==self.N
    
        #if Kss!=None:
            #assert Kss.shape[0]==Ks.shape[1]
            #assert Kss.shape[1]==Ks.shape[1]

        self.vd.getTerm(term_i).getKcf().setK0cross(Ks)


    #TODO: naming. not nice... needs also quite a few more checks, I think.
    def predictMean(self):
        """
        predict the conditional mean (BLUP)

        Returns:
            predictions (BLUP)
        """
        assert self.noisPos!=None,      'No noise element'
        assert self.init,               'GP not initialised'
                
        KiY = self.gp.agetKEffInvYCache()
                
        if self.fast==False:
            KiY = KiY.reshape(self.P,self.N).T
                
        Ymean = None
        for term_i in range(self.n_terms):
            if term_i!=self.noisPos:
                Kstar = self.vd.getTerm(term_i).getKcf().Kcross(SP.zeros(1))
                term  = SP.dot(Kstar.T,KiY)
                if self.P>1:
                    C    = self.getEstTraitCovar(term_i)
                    term = SP.dot(term,C)
                if Ymean==None:     Ymean  = term
                else:               Ymean += term
            
        # to generalise
        Ymean+=self.getFixed()[:,0]
                
        return Ymean





import sys
sys.path.insert(0,'./../../..')
from limix.core.mean import mean

import scipy as SP
import scipy.linalg as LA 
import time as TIME
import copy

if __name__ == "__main__":

    # generate data
    h2 = 0.3
    N = 1000; P = 4; S = 1000
    X = 1.*(SP.rand(N,S)<0.2)
    beta = SP.randn(S,P)
    Yg = SP.dot(X,beta); Yg*=SP.sqrt(h2/Yg.var(0).mean())
    Yn = SP.randn(N,P); Yn*=SP.sqrt((1-h2)/Yn.var(0).mean())
    Y  = Yg+Yn; Y-=Y.mean(0); Y/=Y.std(0)
    XX = SP.dot(X,X.T)
    XX/= XX.diagonal().mean()
    Xr = 1.*(SP.rand(N,10)<0.2)
    Xr*= SP.sqrt(N/(Xr**2).sum())

    # define mean term
    mean = mean(Y)
    print((mean.Y))

    # add first fixed effect
    F = 1.*(SP.rand(N,2)<0.2); A = SP.eye(P)
    mean.addFixedEffect(F=F,A=A)
    # add first fixed effect
    F = 1.*(SP.rand(N,3)<0.2); A = SP.ones((1,P))
    mean.addFixedEffect(F=F,A=A)

    # rotate stuff by row and cols
    C = SP.cov(Y.T)
    Sc,Uc = LA.eigh(C)
    Sr,Ur = LA.eigh(XX)
    d = SP.kron(Sc,Sr)
    mean.d = d
    mean.Lc = Uc.T
    mean.Lr = Ur.T
    mean.LRLdiag = Sr
    mean.LCL     = C**2

    if 1:
        # calculate stuff to see if it goes through
        print((mean.Ystar()))
        print((mean.Yhat()))
        print((mean.Xstar()))
        print((mean.Xhat()))
        print((mean.XstarT_dot(SP.randn(mean.N*mean.P,mean.n_fixed_effs))))
        print((mean.Areml()))
        print((mean.beta_hat()))
        Bhat =  mean.B_hat()
        print((mean.Zstar()))

        # test grad stuff
        print((mean.LRLdiag_Xhat_tens()))
        print((mean.LRLdiag_Yhat()))
        print((mean.Areml_grad()))
        print((mean.beta_grad()))
        print((mean.Xstar_beta_grad()))





# import sys
#
# try:
#     sys.path.insert(0,'./../../..')
#     sys.path.insert(0,'./../../../../../mksum/mtSet')
#     from mtSet.pycore.gp import gp2kronSum as gp2kronSumMtSet
#     import mtSet.pycore.mean as MEAN
#     mtSet_present = True
# except:
#     print("no mtSet found in path")
#     mtSet_present = False
#
# from limix.core.mean import mean
# from limix.core.mean.mean_efficient import Mean
# from limix.core.gp import gp2kronSum as gp2kronSum
#
# from limix.core.covar import freeform
# import limix.core.optimize.optimize_bfgs as OPT
# import limix.core.association.lmm_kronecker as lmm
#
# import scipy as SP
# import scipy.linalg as LA
# import time as TIME
# import copy
#
#
# def generate_data(N=100,S=1000,R=1000,P=2,h2_S=0.1,h2_R=0.0,maf=0.5):
#     """
#     generate SNPs, random effects and phenotypes
#
#     args:
#         N     number individuals
#         S     number SNPs
#         R     number random effects
#         P     number phenotypes
#         h2_S  heritability SNPs
#         h2_R  heritability of random effects
#         maf   minimum allele frequency
#
#     returns:
#         snps  [N x S] ndarray of SNPs
#         X     [N x R] ndarray of random effects
#         Y     [N x P] ndarray of phenotypes
#     """
#
#     assert h2_R>=0, "h2_R has to be greater or equal to zero"
#     assert h2_S>=0, "h2_S has to be greater or equal to zero"
#     assert h2_R+h2_S<=1, "h2_R + h2_S has to be smaller or equal to one"
#
#     X = 1.*(SP.rand(N,R)<0.2)#random effects
#     beta = SP.randn(R,P)
#     Yconf = SP.dot(X,beta);
#     Yconf*=SP.sqrt(h2_R/Yconf.var(0).mean())
#
#     snps = 1.*(SP.rand(N,S)<maf)#SNP effects
#     beta_snp = SP.randn(S,P)*10000
#     Yg = SP.dot(snps,beta_snp);
#     Yg*=SP.sqrt(h2_S/Yg.var(0).mean())
#
#
#     Yn = SP.randn(N,P); Yn*=SP.sqrt((1-h2_S-h2_R))
#     Y  = Yconf+Yg+Yn; Y/=Y.std(0)
#     return snps, X, Y
#
#
# if __name__ == "__main__":
#
#     plot = "plot" in sys.argv
#
#     # generate data
#     N = 500; P = 2; S = 1000; R = 1000
#     maf = 0.2
#     h2_S = 0.3; h2_R = 0.2
#     snps, X, Y = generate_data(N=N,S=S,R=R,P=P,h2_S=h2_S,h2_R=h2_R,maf=maf)
#
#     # define mean term
#     mu = mean(Y)
#
#     # add any fixed effect including bias term
#     F = 1.*(SP.rand(N,3)<0.2); A = SP.eye(P)
#     F[:,0]=1.0
#     mu.addFixedEffect(F=F,A=A)
#
#     # add common fixed effect
#     F2 = 1.*(SP.rand(N,4)<0.2); A2 = SP.ones((1,P))
#     mu.addFixedEffect(F=F2,A=A2)
#
#
#     # define covariance matrices
#     Cg = freeform(P)
#     Cn = freeform(P)
#     Cg1 = freeform(P)
#     Cn1 = freeform(P)
#
#     # compare with mtSet implementation
#     params = {}
#     params['Cg']   = SP.randn(int(0.5*P*(P+1)))
#     params['Cn']   = SP.randn(int(0.5*P*(P+1)))
#
#
#     print("creating gp2kronSum object")
#     XX = SP.dot(X,X.T)
#     XX/= XX.diagonal().mean()
#     gp = gp2kronSum(mu,Cg,Cn,XX)
#     gp.setParams(params)
#
#     if "ML" in sys.argv:
#         print("ML estimation")
#         gp.set_reml(False)
#     else:
#         print("REML estimation")
#
#     print("optimization of GP parameters")
#     start = TIME.time()
#     conv,info = OPT.opt_hyper(gp,params,factr=1e3)
#     print(('time for fitting GP:', TIME.time()-start))
#
#
#     print(conv)
#
#     print("creating lmm for association using GP object")
#     assoc = lmm.LmmKronecker(gp=gp)
#
#     #test snps
#     print("testing SNPs with any effect")
#     pv,LL_snps,LL_snps_0= assoc.test_snps(snps)
#
#
#
#     if 1:
#         print("forward selection step")
#         print("adding SNP with smalles pv as fixed effect")
#         i_pv = pv.argsort()
#         assoc.addFixedEffect(F=snps[:,i_pv[0]:(i_pv[0]+1)],A=None)
#
#         if 1:#optimize
#             print("refitting GP")
#             start = TIME.time()
#             conv,info = OPT.opt_hyper(assoc._gp,params,factr=1e3)
#             print(('time for fitting GP:', TIME.time()-start))
#
#
#         print(conv)
#
#         print("testing after forward selection step")
#         assoc_forw = lmm.LmmKronecker(gp=assoc._gp)
#         pv_forw,LL_snps_forw,LL_snps_0_forw = assoc_forw.test_snps(snps)
#
#
#         if plot:
#             import pylab as pl
#             pl.ion()
#             print("plotting pvalues before vs after forward selection")
#             pl.figure()
#             pl.plot(-SP.log10(pv_forw),-SP.log10(pv),'.')
#             pl.plot([0,8],[0,8])



__author__ = 'damienarnol1'

import numpy as np
import h5py
import scipy as sp
from limix.utils.preprocess import covar_rescaling_factor


def make_sure_reasonable_conditioning(S):
    max_cond = 1e1
    cond = S.max() / S.min()
    if cond > max_cond:
        eps = (max_cond * S.min() - S.max()) / (1 - max_cond)
        # logger = logging.getLogger(__name__)
        # logger.warn("The covariance matrix's conditioning number" +
        #             " is too high: %e. Summing %e to its eigenvalues and " +
        #             "renormalizing for a better conditioning number.",
        #             cond, eps)
        m = S.mean()
        S += eps
        S *= m / S.mean()
        print('S changed')
    else:
        print("S unchanged ")
    return S


def _generate_SE(N_cells, X=None, off=False):
    if X is None:
        X = np.random.uniform(0, N_cells * 2, [N_cells, 2])
    distance = np.empty([N_cells, N_cells])

    scale = 0 if off else np.random.uniform(0.5, 4, 1)
    length = np.random.uniform(N_cells / 7.0, N_cells / 6.0, 1)

    for i in range(0, N_cells):
        for j in range(0, N_cells):
            tmp = X[i, :] - X[j, :]
            distance[i, j] = tmp[0] ** 2 + tmp[1] ** 2

    cov = scale * sp.exp(-distance / (2 * length))
    # cov *= covar_rescaling_factor(cov)
    return cov, X, scale, length


def normalise_columns(mat):
    mat -= mat.mean(axis=0)
    mat /= mat.std(axis=0)
    mat /= np.sqrt(mat.shape[1])
    return mat


def generate_kinship(N_cells):
    tmp = np.random.randn(N_cells, 2 * N_cells)
    # tmp = normalise_columns(tmp)
    K = tmp.dot(tmp.transpose())
    K *= covar_rescaling_factor(K)
    return K


def generate_zkz(N_cells, rm_diag=False, Kinship=None, X=None, off=False):
    se_all = _generate_SE(N_cells, X=X, off=off)
    se = se_all[0]
    if rm_diag:
        se -= np.eye(se.shape[0]) * se.diagonal()
    se_all = np.delete(se_all, 0)
    if Kinship is None:
        Kinship = generate_kinship(N_cells)
    zkz = se.dot(Kinship.dot(se.transpose()))
    return zkz, se_all, Kinship


def generate_complete_model(N_cells, X=None, Kinship=None,  switch_off=[], rm_diagonal=False):

    # direct effect term
    direct_scale = 0 if 'direct' in switch_off else np.random.uniform(0.5, 4, 1)
    if Kinship is None:
        Kinship = generate_kinship(N_cells)
    direct_cov = direct_scale * Kinship

    # social/environment effect term
    off = 'social' in switch_off
    zkz, social_params, Kinship = generate_zkz(N_cells, rm_diag=rm_diagonal, Kinship=Kinship, X=X, off=off)

    # global noise
    noise_scale = 0 if 'noise' in switch_off else np.random.uniform(0.5, 4, 1)
    noise = noise_scale * np.eye(N_cells)

    # local noise
    off = 'local' in switch_off
    local_noise_cov, X, local_noise_scale, local_noise_length = _generate_SE(N_cells,
                                                                             X=social_params[0],
                                                                             off=off)

    # total cov and mean
    cov = local_noise_cov + noise + direct_cov + zkz
    mean = np.array([0] * N_cells)

    data = np.random.multivariate_normal(mean, cov, 1).transpose()

    return data, {'direct_scale': direct_scale,
                  'noise_scale': noise_scale,
                  'local_noise_scale': local_noise_scale, 'local_noise_length': local_noise_length,
                  'social_scale': social_params[1], 'social_length': social_params[2],
                  'Kinship': Kinship, 'X': X}


def generate_simple(N_cells, rm_zkz_diag=False):
    # one ZKZ matrix
    zkz, se_params, Kinship = generate_zkz(N_cells, rm_diag=rm_zkz_diag)

    # one noise matrix
    sigma_noise = np.random.uniform(0.5, 1.5, 1)
    noise = sigma_noise * np.eye(N_cells)

    # grouping parameters
    parameters = {'X': se_params[0], 'scale': se_params[1], 'length': se_params[2], 'Kinship': Kinship,
                  'noise': sigma_noise}

    # generating data
    covar = zkz + noise
    mean = np.array([0] * N_cells)
    data = np.random.multivariate_normal(mean, covar, 1).transpose()

    return data, parameters


def generate_samples(N_cells):
    # Number of cells
    N_cells = 500

    # model parameters
    sigma_I = 2
    sigma_E = 4
    sigma_eps = 1
    sigma_epsLocal = 1
    length_E = 100
    length_eps = 200

    # covariances structure
    # Kinship
    # TODO come up with a less random kinship matrix
    tmp = np.random.randn(N_cells, 2.0 * N_cells)
    Kinship = tmp.dot(tmp.T)
    Kinship = Kinship / np.mean(Kinship.diagonal())
    # Kinship *= covar_rescaling_factor(Kinship)

    # Z matrix
    X = np.random.uniform(0, N_cells * 2, [N_cells, 2])
    distance = np.empty([N_cells, N_cells])
    for i in range(0, N_cells):
        for j in range(0, N_cells):
            tmp = X[i, :] - X[j, :]
            distance[i, j] = tmp[0] ** 2 + tmp[1] ** 2

    Z_E = sigma_E ** 2 * sp.exp(-distance / 2 * length_E)
    Z_eps = sigma_eps ** 2 * sp.exp(-distance / 2 * length_eps)

    # total covariance
    covar = sigma_I ** 2 * Kinship + \
            Z_E.dot(Kinship.dot(Z_E.transpose())) + \
            Z_eps + \
            sigma_eps ** 2 * np.eye(N_cells)

    # mean
    mean = np.array([0] * N_cells)

    # generate random phenotype
    phenotype = np.random.multivariate_normal(mean, covar, 2)
    phenotype = phenotype.transpose()

    return Kinship, phenotype



__author__ = 'damienarnol1'

import numpy as np
import os
import glob
from . import util_functions
import shutil



def split_expression_file(expression_file, IMC_dir, output_dir):
    # read protein names
    with open(expression_file, 'r') as f:
        prot_tmp = f.readline()
    protein_names = prot_tmp.split(' ')
    protein_names[-1] = protein_names[-1][0:-1]  # removing the newline sign at the end of the last protein
    protein_names = protein_names[1:]
    protein_names = ' '.join(protein_names)

    # protein_names = np.reshape(protein_names, [len(protein_names), 1])

    phenotypes = np.loadtxt(expression_file, delimiter=' ', skiprows=1)
    image_names = sorted(os.listdir(IMC_dir))

    for image_index in range(0, len(image_names)):
        phenotype = phenotypes[phenotypes[:, 0] == image_index, 1:]
        file_name = image_names[image_index].split('_')[0]
        full_path = output_dir + '/' + file_name

        with open(full_path, 'w') as f:
            np.savetxt(f, phenotype, delimiter=' ', header=protein_names, comments='')


def create_analysis_tree(position_dir, expressions_dir, analysis_dir):
    # name of the images to check positions and expressions match
    position_names = os.listdir(position_dir)
    position_names = [name.split('_')[0] for name in position_names]
    position_names = sorted(position_names)

    expression_names = os.listdir(expressions_dir)
    expression_names = sorted(expression_names)

    # full path of the file for processing
    position_files = glob.glob(position_dir+'/*')
    position_files = sorted(position_files)

    expression_files = glob.glob(expressions_dir+'/*')
    expression_files = sorted(expression_files)

    # reading and creating tree
    for image_index in range(0, len(position_names)):
        if position_names[image_index] != expression_names[image_index]:
            print(position_names[image_index] + 'is not equal to ' + expression_names[image_index])
            raise Exception("Image names dont match for position and expression data")

        # create image directory
        dir_name = analysis_dir+'/'+position_names[image_index]
        dir_name = util_functions.make_dir(dir_name)

        # create file names
        position_file_cp = dir_name + '/' + 'positions.txt'
        expression_file_cp = dir_name + '/' + 'expressions.txt'

        # copy files
        shutil.copy(position_files[image_index], position_file_cp)
        shutil.copy(expression_files[image_index], expression_file_cp)


# 1 - define necessary file names and directory names
# 2 - create analysis tree
# 3 - submit one job per directory of the tree

all_expressions_file = '/homes/arnol/random_effect_model/data/membrane_expressions'
IMC_dir='/hps/nobackup/stegle/users/arnol/data/SampleSet_1/IMC_data/'
expression_dir = '/homes/arnol/random_effect_model/data/expression_levels/'
positions_dir = '/homes/arnol/random_effect_model/data/positions/'
analysis_dir = '/homes/arnol/random_effect_model/analysis_2/'

split_expression_file(all_expressions_file, IMC_dir, expression_dir)
create_analysis_tree(positions_dir, expression_dir, analysis_dir)

# protein list
protein_list = ['pAKT', 'AMPK', 'pBAD', 'bcatenin', 'CAHIX', 'CD20', 'CD3', 'CD44', 'CD45', 'CD68', 'CC3', 'cMyc',
                'CREB', 'Cytokeratin7', 'PanKeratin', 'Ecadherin', 'EGFR', 'EpCAM', 'ERa', 'pERK12', 'Fibronectin',
                'GATA3', 'HER2', 'H3', 'Ki67', 'PRAB', 'S6', 'SHP2', 'Slug', 'SMA', 'Twist', 'Vimentin']

# running model on every directory

# list directories
image_dirs = glob.glob(analysis_dir+'/*')
for image_dir in image_dirs:
    result_directory = image_dir+'/results/'
    util_functions.make_dir(result_directory, overwrite=False)
    for protein_1 in protein_list:
        for protein_2 in protein_list:
            file_name = result_directory+'/'+protein_1+'_'+protein_2
            command_line = \
                'bsub -o /homes/arnol/random_effect_model/log -M 15000 -R "rusage[mem=15000]" python pairwise_interaction.py ' + \
                protein_1 + ' ' + protein_2 + ' ' + file_name + ' ' + image_dir
            os.system(command_line)






import scipy as sp
import pdb


def sq_dist(X1,X2=None):
    """
    computes a matrix of all pariwise squared distances
    """
    if X2==None:
        X2 = X1
    assert X1.shape[1]==X2.shape[1], 'dimensions do not match'

    n = X1.shape[0]
    m = X2.shape[0]
    d = X1.shape[1]
    # (X1 - X2)**2 = X1**2 + X2**2 - 2X1X2
    X1sq = sp.reshape((X1**2).sum(1),n,1)
    X2sq = sp.reshape((X2**2).sum(1),m,1)

    K = sp.tile((X1*X1).sum(1),(m,1)).T + sp.tile((X2*X2).sum(1),(n,1)) - 2*sp.dot(X1,X2.T)
    return K



# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
#Licensed under the Apache License, Version 2.0 (the "License");
#you may not use this file except in compliance with the License.
#You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
#Unless required by applicable law or agreed to in writing, software
#distributed under the License is distributed on an "AS IS" BASIS,
#WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#See the License for the specific language governing permissions and
#limitations under the License.

__all__ = ['']

#make core available on flat import
from .core import *
#verbose flags for the entire module
_verbose = False


def getVerbose(verbose=None):
	"""resolve verbose flag, using module settings if verbose=None"""
	global _verbose
	if verbose is None:
		return _verbose
	_verbose = verbose
	return verbose



import numpy as np
import scipy as sp
import scipy.stats
from numpy import dot
from limix.core.linalg.linalg_matrix import QS_from_K
import sys

def standardize_design(G, mean_var=None):
    if mean_var is None:
        mean_var = (0., 1./G.shape[1])
        np.apply_along_axis(lambda x: _change_sample_stats(x, (0., 1.)), 0, G)
    else:
        G -= mean_var[0]
        G /= np.sqrt(mean_var[1])

def standardize_cov(K):
    from limix.utils.preprocess import covar_rescale
    return covar_rescale(K)

def static_effsiz_sampler(effects):
    def sampler(n):
        return effects
    return sampler

def normal_effsiz_sampler():
    def sampler(n):
        if n == 0:
            return np.empty(0, float)
        effects = np.random.randn(n)
        return effects
    return sampler

def binary_effsiz_sampler():
    def sampler(n):
        if n == 0:
            return np.empty(0, float)
        effects = np.random.randn(n)
        effects[effects >= 0] = +1.
        effects[effects <  0] = -1.
        return effects
    return sampler

def geometric_ntrials_sampler(min_ntrials, max_ntrials, p):
    def sampler():
        k = np.random.geometric(p) + min_ntrials - 1
        while k >= max_ntrials:
            k = np.random.geometric(p) + min_ntrials - 1
        return k
    return sampler

class TraitSampler(object):
    def __init__(self):
        self._Gs = dict()
        self._zc = dict()
        self._causal_indices = dict()
        self._us = dict()
        self._eff_sample_mean_vars = dict()
        self._noise_var = 0.
        self._nindividuals = None

    def add_effects_design(self, name, G, effsiz_sampler,
                           effsiz_sample_mean_var=None,
                           eff_sample_mean_var=None,
                           ncausal=None,
                           min_dist=1,
                           base_pos=None):

        if ncausal is None:
            ncausal = G.shape[1]

        self._Gs[name] = G

        if base_pos is None:
            causal_indices = self._sample_causal_indices(G.shape[1], ncausal)
        else:
            causal_indices = self._sample_causal_indices_pos(base_pos, ncausal,
                                                             min_dist)
        self._causal_indices[name] = causal_indices

        u = effsiz_sampler(len(causal_indices))

        if effsiz_sample_mean_var is not None:
            (m, v) = effsiz_sample_mean_var
            _change_sample_stats(u, mean_var=(m, v))

        self._us[name] = u
        self._eff_sample_mean_vars[name] = eff_sample_mean_var

        if self._nindividuals is None:
            self._nindividuals = G.shape[0]

    def causal(self, name):
        return self._causal_indices[name]

    def add_effect_cov(self, name, K=None, effsize_sample_mean_var=None,
                       Q=None, S=None):
        import scipy as sp
        import scipy.stats
        if Q is None or S is None:
            (Q, S) = QS_from_K(K)
        S = np.sqrt(S)

        u = np.random.randn(S.shape[0])
        if effsize_sample_mean_var is not None:
            m = effsize_sample_mean_var[0]
            v = effsize_sample_mean_var[1]
            _change_sample_stats(u, (m, v))

        self._zc[name] = dot(Q, S * u)

        if self._nindividuals is None:
            self._nindividuals = Q.shape[0]

    def set_noise(self, var):
        self._noise_var = var

    def get_noise(self):
        return self._noise_var

    def sample_traits(self):
        z = 0
        n = self._nindividuals
        ze = np.random.randn(n) * np.sqrt(self._noise_var)
        z = ze.copy()
        zd = dict()
        for effect_name in list(self._Gs.keys()):
            G = self._Gs[effect_name]
            u = self._us[effect_name]
            idx = self._causal_indices[effect_name]
            zd[effect_name] = dot(G[:, idx], u)

            if self._eff_sample_mean_vars[effect_name] is not None:
                zd[effect_name] = _change_sample_stats(zd[effect_name],
                                               self._eff_sample_mean_vars)

            z += zd[effect_name]

        for z_ in list(self._zc.values()):
            z += z_

        return (z, zd, self._zc, ze)

    def _sample_causal_indices_trial(self, base_pos, ncausal, min_dist=1):
        size = len(base_pos)
        causals = []

        failed = 0
        while len(causals) < ncausal and failed < ncausal * 100:
            i = np.random.randint(size)
            if self._distance_next_causal(i, causals, base_pos) >= min_dist:
                causals.append(i)
            else:
                failed += 1

        if len(causals) != ncausal:
            return None

        return np.array(causals, int)

    def _sample_causal_indices_pos(self, base_pos, ncausal, min_dist=1):
        while True:
            causals = self._sample_causal_indices_trial(base_pos, ncausal,
                                                        min_dist)
            if causals is not None:
                return causals

    def _sample_causal_indices(self, size, ncausal):
        return np.random.choice(size, ncausal, replace=False)

    def _distance_next_causal(self, causal, causals, base_pos):
        if len(causals) == 0:
            return np.inf
        pos = base_pos[causal]
        poss = base_pos[causals]
        return np.min(np.abs(poss - pos))

    def _zmean(self):
        assert self.get_noise() == 0.
        (z, _, _, ze) = TraitSampler.sample_traits(self)
        return np.mean(z - ze)

def _change_sample_stats(x, mean_var=(None, None)):
    if mean_var[0] is not None:
        x -= np.mean(x)
        x += mean_var[0]

    if mean_var[1] is not None:
        v0 = np.mean(x**2)
        v1 = mean_var[1]

        if v0 == 0.:
            x[:] = np.sqrt(v1)
        else:
            c = np.sqrt(v1/v0)
            x *= c

class BernoulliTraitSampler(TraitSampler):
    def __init__(self):
        TraitSampler.__init__(self)

    def _offset_due_prevalence(self, var_noise, prevalence):

        print ("Calculating offset due to prevalence...")

        (_, z) = self._sample_traits_once(var_noise, 0.)
        offset = np.percentile(z, 100. * (1-prevalence))
        print(("Calculated offset: %.3f" % offset))
        return offset

    def _sample_traits_once(self, var_noise, offset):

        (z, _, _, _) = TraitSampler.sample_traits(self)
        z += np.random.randn(z.shape[0]) * np.sqrt(var_noise)
        y = np.zeros(z.shape[0], float)
        y[z >= offset] = 1.

        return (y, z)

    def sample_traits(self, pop_size, var_noise, prevalence=0.5, ascertainment=0.5):
        print(("Prevalence: %.3f." % prevalence))
        print(("Ascertainment: %.3f." % ascertainment))
        offset = self._offset_due_prevalence(var_noise, prevalence)

        print("Sampling traits...")
        n1 = int(ascertainment * pop_size)
        n0 = int(pop_size) - n1

        (y_, z_) = self._sample_traits_once(var_noise, offset)
        ok0 = np.where(y_ == 0.)[0]
        ok1 = np.where(y_ == 1.)[0]

        assert len(ok0) >= n0
        assert len(ok1) >= n1

        np.random.shuffle(ok0)
        np.random.shuffle(ok1)

        ok0 = ok0[:n0]
        ok1 = ok1[:n1]

        ok = np.concatenate((ok0, ok1))

        y = y_[ok]
        z = z_[ok]

        selected_individuals = ok

        print("Done.")
        return (y, z, offset, selected_individuals)

class BinomialTraitSampler(TraitSampler):
    def __init__(self):
        TraitSampler.__init__(self)

    def _sample_ntrials(self, pop_size, ntrials_sampler):
        ntrials = []
        for i in range(pop_size):
            ntrials.append(ntrials_sampler())
        return np.array(ntrials, int)

    def sample_traits(self, pop_size, vare, ntrials_sampler):

        print("Sampling traits...")

        (z, _, _, _) = TraitSampler.sample_traits(self)
        selected_individuals = np.random.choice(len(z), pop_size, replace=False)
        z = z[selected_individuals]

        ntrials = self._sample_ntrials(pop_size, ntrials_sampler)

        E = np.random.randn(pop_size, np.max(ntrials))
        E *= np.sqrt(vare)

        Z = z[:, np.newaxis] + E

        Z[Z >  0.] = 1.
        Z[Z <= 0.] = 0.

        y = np.empty(pop_size)
        for i in range(y.shape[0]):
            y[i] = np.sum(Z[i,:ntrials[i]])

        print("Done.")

        y = dict(trait=y, ntrials=ntrials)
        return (y, selected_individuals)


if __name__ == '__main__':
    pass



from .mvSet import MvSetTest
from .mvSetFull import MvSetTestFull
from .mvSetInc import MvSetTestInc
import pandas as pd

def fit_iSet(Y=None, Xr=None, F=None, Rr=None, factr=1e7, Rg=None, Ug=None, Sg=None, Ie=None, n_nulls=10):
    # data
    noneNone = Sg is not None and Ug is not None
    bgRE = Rg is not None or noneNone
    # fixed effect
    msg = 'The current implementation of the full rank iSet'
    msg+= ' does not support covariates.'
    msg+= ' We reccommend to regress out covariates and'
    msg+= ' subsequently quantile normalize the phenotypes'
    msg+= ' to a normal distribution prior to use mtSet/iSet.'
    msg+= ' This can be done within the LIMIX framework using'
    msg+= ' the methods limix.utils.preprocess.regressOut and'
    msg+= ' limix.utils.preprocess.gaussianize'
    assert not (F is not None and bgRE), msg
    # strat
    strat = Ie is not None
    msg = 'iSet for interaction analysis of stratified populations ' 
    msg+= 'using contextual variables does not support random effect '
    msg+= 'correction for confounding. '
    msg+= 'Please use the fixed effects to correct for confounding. '
    assert not (strat and bgRE), msg

    #define mtSet
    if bgRE:        mvset = MvSetTestFull(Y=Y,Xr=Xr,Rg=Rg,Ug=Ug,Sg=Sg,factr=factr)
    elif strat:     mvset = MvSetTestInc(Y=Y,Xr=Xr,F=F,Ie=Ie,factr=factr)
    else:           mvset = MvSetTest(Y=Y,Xr=Xr,F=F,Rr=Rr,factr=factr)

    RV = {}
    RV['mtSet LLR'] = mvset.assoc()
    RV['iSet LLR'] = mvset.gxe()
    RV['iSet-het LLR'] = mvset.gxehet()
    RV['Persistent Var'] = mvset.info['full']['var_r'][0]
    RV['Rescaling-GxC Var'] = mvset.info['full']['var_r'][1]
    RV['Heterogeneity-GxC var'] = mvset.info['full']['var_r'][2]
    df = pd.DataFrame(RV)

    RV0 = {}
    RV0['mtSet LLR0'] = mvset.assoc_null(n_nulls=n_nulls)
    RV0['iSet LLR0'] = mvset.gxe_null(n_nulls=n_nulls)
    RV0['iSet-het LLR0'] = mvset.gxehet_null(n_nulls=n_nulls)
    df0 = pd.DataFrame(RV0)

    return df, df0




"""Variance Decomposition testing code"""
import unittest
import scipy as SP
import scipy.stats
import pdb
import os
import sys
import limix
import limix.deprecated.modules.varianceDecomposition as VAR
from limix.test import data


class VarianceDecompoitionKronecker_test(unittest.TestCase):
    """test class for VarianceDecomposition
         THis testing class assumes everything is Kronecker.
    """

    def genGeno(self):
        X  = (SP.rand(self.N,self.S)<0.2)*1.
        self.D['X'] = X


    def genPheno(self):
        dp = SP.ones(self.P); dp[1]=-1
        Y = SP.zeros((self.N,self.P))
        gamma0_g = SP.randn(self.S)
        gamma0_n = SP.randn(self.N)
        for p in range(self.P):
            gamma_g = SP.randn(self.S)
            gamma_n = SP.randn(self.N)
            beta_g  = dp[p]*gamma0_g+gamma_g
            beta_n  = gamma0_n+gamma_n
            y=SP.dot(self.D['X'],beta_g)
            y+=beta_n
            Y[:,p]=y
        Y=SP.stats.zscore(Y,0)
        self.D['Y']= Y


    def setUp(self):
        #check: do we have a csv File?
        self.dir_name = os.path.dirname(__file__)
        self.dataset = os.path.join(self.dir_name,'varDecomp')

        if (not os.path.exists(self.dataset)) or 'recalc' in sys.argv:
            if not os.path.exists(self.dataset):
                os.makedirs(self.dataset)
            SP.random.seed(1)
            self.N = 200
            self.S = 1000
            self.P = 2
            self.D = {}
            self.genGeno()
            self.genPheno()
            self.generate = True
        else:
            self.generate=False
            #self.D = data.load(os.path.join(self.dir_name,self.dataset))
            self.D = data.load(self.dataset)
            self.N = self.D['X'].shape[0]
            self.S = self.D['X'].shape[1]
            self.P = self.D['Y'].shape[1]

        self.Kg = SP.dot(self.D['X'],self.D['X'].T)
        self.Kg = self.Kg/self.Kg.diagonal().mean()

        self.vc = VAR.VarianceDecomposition(self.D['Y'])
        self.vc.addRandomEffect(self.Kg,jitter=0)
        self.vc.addRandomEffect(is_noise=True,jitter=0)
        self.vc.addFixedEffect()

    def test_fit(self):
        """ optimization test """
        self.vc.optimize(verbose=False)
        params = self.vc.getScales()
        if self.generate:
            self.D['params_true'] = params
            data.dump(self.D,self.dataset)
            self.generate=False
        params_true = self.D['params_true']
        RV = ((SP.absolute(params)-SP.absolute(params_true))**2).max()
        self.assertTrue(RV<1e-6)

    def test_fitFast(self):
        """ optimization test """
        self.vc.optimize(fast=True,verbose=False)
        params = self.vc.getScales()
        if self.generate:
            self.D['params_true'] = params
            data.dump(self.D,self.dataset)
            self.generate=False

        params_true = self.D['params_true']
        #make sign invariant
        RV = ((SP.absolute(params)-SP.absolute(params_true))**2).max()<1e-6
        self.assertTrue(RV)

class VarianceDecompositionSoftKronecker_test(unittest.TestCase):
    """test class for VarianceDecomposition
         THis testing class assumes tests the ability to
    """

    def genGeno(self):
        X  = (SP.rand(self.N,self.S)<0.2)*1.
        self.D['X'] = X

    def genPheno(self):
        dp = SP.ones(self.P); dp[1]=-1
        Y = SP.zeros((self.N,self.P))
        gamma0_g = SP.randn(self.S)
        gamma0_n = SP.randn(self.N)
        for p in range(self.P):
            gamma_g = SP.randn(self.S)
            gamma_n = SP.randn(self.N)
            beta_g  = dp[p]*gamma0_g+gamma_g
            beta_n  = gamma0_n+gamma_n
            y=SP.dot(self.D['X'],beta_g)
            y+=beta_n
            Y[:,p]=y
        Y=SP.stats.zscore(Y,0)
        self.D['Y']= Y

    def setUp(self):
        #check: do we have a csv File?
        self.dir_name = os.path.dirname(__file__)
        self.dataset = os.path.join(self.dir_name,'varDecomp')

        if (not os.path.exists(self.dataset)) or 'recalc' in sys.argv:
            if not os.path.exists(self.dataset):
                os.makedirs(self.dataset)
            SP.random.seed(1)
            self.N = 200
            self.S = 1000
            self.P = 2
            self.D = {}
            self.genGeno()
            self.genPheno()
            self.generate = True
        else:
            self.generate=False
            #self.D = data.load(os.path.join(self.dir_name,self.dataset))
            self.D = data.load(self.dataset)
            self.N = self.D['X'].shape[0]
            self.S = self.D['X'].shape[1]
            self.P = self.D['Y'].shape[1]

        self.Kg = SP.dot(self.D['X'],self.D['X'].T)
        self.Kg = self.Kg/self.Kg.diagonal().mean()

        #add missing values to Y
        if 1:
            self.D['Y'][0,0] = SP.nan
            self.D['Y'][1,1] = SP.nan
            self.D['Y'][10,0] = SP.nan
            self.D['Y'][100,1] = SP.nan
        self.vc = VAR.VarianceDecomposition(self.D['Y'])
        self.vc.addRandomEffect(self.Kg,jitter=0)
        self.vc.addRandomEffect(is_noise=True,jitter=0)
        self.vc.addFixedEffect()

    def test_fit(self):
        """ optimization test """
        self.vc.optimize(verbose=False)
        params = self.vc.getScales()
        if self.generate:
            self.D['params_true'] = params
            data.dump(self.D,self.dataset)
            self.generate=False
        params_true = self.D['params_true']
        RV = ((SP.absolute(params)-SP.absolute(params_true))**2).max()
        #permit more flexibility, as we set a few values to NAN
        self.assertTrue(RV<1e-4)


if __name__ == '__main__':
    unittest.main()



import scipy as sp
import scipy.linalg as LA

def dS_dti(dM_dti,M=None,U=None,S=None):
    assert M is not None or U is not None, 'Specify either M or U and S'
    if U is None:
        U,S = LA.eigh(M) 
    RV = (U*sp.dot(dM_dti,U)).sum(0)
    return RV
    
def dU_dti(dM_dti,M=None,U=None,S=None):
    assert M is not None or (U is not None and S is not None), 'Specify either M or U and S'
    if U is None or S is None:
        U,S = LA.eigh(M) 
    P  = dM_dti.shape[0]
    RV = sp.zeros((P,P))
    for p in range(P):
        S1 = S[p]-S
        I  = abs(S1)>1e-4
        _U = U[:,I]
        _D = (S1[I]**(-1))[:,sp.newaxis]
        pseudo = sp.dot(_U,_D*_U.T)
        RV[:,p] = sp.dot(pseudo,sp.dot(dM_dti,U[:,p])) 
    return RV




# Copyright(c) 2014, The LIMIX developers (Christoph Lippert, Paolo Francesco Casale, Oliver Stegle)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import scipy as sp 
import pandas as pd


try:
    #see if fastlmm is in the path for the fast C-based parser
    import fastlmm.pyplink.wrap_plink_parser as parser  
    cparser=True
except:
    cparser=False
    pass

def estCumPos(position,offset=0,chrom_len=None):
    '''
    compute the cumulative position of each variant given the position and the chromosome
    Also return the starting cumulativeposition of each chromosome

    Args:
        position:   pandas DataFrame of basepair positions (key='pos') and chromosome values (key='chrom')
                    The DataFrame will be updated with field 'pos_cum'
        chrom_len:  vector with predefined chromosome length
        offset:     offset between chromosomes for cumulative position (default 0 bp)
    
    Returns:
        chrom_pos,position:
        chrom_pos:  numpy.array of starting cumulative positions for each chromosome
        position:   augmented position object where cumulative positions are defined
    '''
    RV = position.copy()
    chromvals =  sp.unique(position['chrom'])# sp.unique is always sorted
    chrom_pos_cum= sp.zeros_like(chromvals)#get the starting position of each Chrom
    pos_cum= sp.zeros_like(position.shape[0])
    if not 'pos_cum' in position:
        RV["pos_cum"]= sp.zeros_like(position['pos'])#get the cum_pos of each variant.
    pos_cum=RV['pos_cum'].values
    maxpos_cum=0
    for i,mychrom in enumerate(chromvals):
        chrom_pos_cum[i] = maxpos_cum
        i_chr=position['chrom']==mychrom
        if chrom_len is None:
            maxpos = position['pos'][i_chr].max()+offset
        else:
            maxpos = chrom_len[i]+offset
        pos_cum[i_chr.values]=maxpos_cum+position.loc[i_chr,'pos']
        maxpos_cum+=maxpos      
    
    return RV,chrom_pos_cum
        
        
def imputeMissing(X, center=True, unit=True, betaNotUnitVariance=False, betaA=1.0, betaB=1.0):
    '''
    fill in missing values in the SNP matrix by the mean value
    optionally center the data and unit-variance it

    Args:
        X:      scipy.array of SNP values. If dtype=='int8' the missing values are -9, 
                otherwise the missing values are scipy.nan
        center: Boolean indicator if data should be mean centered
                Not supported in C-based parser
        unit:   Boolean indicator if data should be normalized to have unit variance
                Not supported in C-based parser
        betaNotUnitVariance:    use Beta(betaA,betaB) standardization instead of unit variance 
                                (only with C-based parser) (default: False)
        betaA:  shape parameter for Beta(betaA,betaB) standardization (only with C-based parser)
        betaB:  scale parameter for Beta(betaA,betaB) standardization (only with C-based parser)
    
    Returns:
        X:      scipy.array of standardized SNPs with scipy.float64 values
    '''
    typeX=X.dtype
    if typeX!= sp.int8:
        iNanX = X!=X
    else:
        iNanX = X==-9
    if iNanX.any() or betaNotUnitVariance:
        if cparser and center and (unit or betaNotUnitVariance):
            print("using C-based imputer")
            if X.flags["C_CONTIGUOUS"] and typeX== sp.float32:
                parser.standardizefloatCAAA(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
                X= sp.array(X,dtype= sp.float64)
            elif X.flags["C_CONTIGUOUS"] and typeX== sp.float64:
                parser.standardizedoubleCAAA(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
            elif X.flags["F_CONTIGUOUS"] and typeX== sp.float32:
                parser.standardizefloatFAAA(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
                X= sp.array(X,dtype= sp.float64)
            elif X.flags["F_CONTIGUOUS"] and typeX== sp.float64:
                parser.standardizedoubleFAAA(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
            else:
                X= sp.array(X,order="F",dtype= sp.float64)
                X[iNanX]= sp.nan
                parser.standardizedoubleFAAA(X,betaNotUnitVariance=betaNotUnitVariance,betaA=betaA,betaB=betaB)
        elif betaNotUnitVariance:
                raise NotImplementedError("Beta(betaA,betaB) standardization only in C-based parser, but not found")
        else:
            nObsX = (~iNanX).sum(0)
            if typeX!= sp.float64:
                X= sp.array(X,dtype= sp.float64)
            X[iNanX] = 0.0
            sumX = (X).sum(0)                
            meanX = sumX/nObsX
            if center:
                X-=meanX
                X[iNanX] = 0.0
                X_=X
            else:
                mean= sp.tile(meanX,(X.shape[0],1))
                X[iNanX]=mean[iNanX]
                X_=X-mean
            if unit:
                stdX =  sp.sqrt((X_*X_).sum(0)/nObsX)
                stdX[stdX==0.0]=1.0
                X/=stdX
    else:
        if X.dtype!= sp.float64:
            X= sp.array(X,dtype= sp.float64)
        if center:
            X-= X.mean(axis=0)
        if unit:
            stdX= X.std(axis=0)
            stdX[stdX==0.0]=1.0
            X/=stdX
    return X


def merge_indices(indices,header=None,join="inner"):
    """
    returns a merged index

    Args:
        indices:    list of indices (e.g. individual IDs)
        header:     list with name of each element of indices (e.g. ["pheno","geno","covars"])
        join:       type of join to perform (standard is "inner"

    Returns:
        pandas DataFrame with merged indices
    """
    indexpd = []
    for i, index in enumerate(indices):
        if header is None:
            header_=[i]
        else:
            header_=[header[i]]
        indexpd.append(pd.DataFrame(data= sp.arange(len(index)),index=index,columns=header_) )
    ret = pd.concat(objs=indexpd, axis=1, join=join)
    return ret

if __name__ == "__main__":
    lists=[["a","b"],["a","c","b"],["d","a","b"]]
    header = [["bl"],["n"],["s"]]
    merge=merge_indices(lists, header=None, join="outer")
    


class TooExpensiveOperationError(RuntimeError):
    def __init__(self, msg):
        RuntimeError.__init__(self, msg)

class NotArrayConvertibleError(TypeError):
    def __init__(self, msg):
        TypeError.__init__(self, msg)

class UndefinedInputError(RuntimeError):
    def __init__(self, msg):
        RuntimeError.__init__(self, msg)

if __name__ == '__main__':

    print(('Oi danilo %d'
          'Como voce esta? %d' % (1, 2)))

    def expensive():
        raise TooExpensiveOperationError('ola')

    try:
        expensive()
    except TooExpensiveOperationError as e:
        print(e)



# __author__ = 'damienarnol1'
#
# # THIS IS FAILING. FIX IT!!!!!!!!!!!!!!!!!!!!!!!!!
# 
# import pdb
#
# import numpy as np
#
# from limix.core.covar import SQExpCov
# from limix.core.covar import ZKZCov
# from limix.core.covar import FixedCov
# from limix.core.covar import SumCov
# from limix.core.gp import GP
# from limix.core.mean import MeanBase
# from limix.utils.preprocess import covar_rescaling_factor
#
# from limix.IMC import generate_data
#
#
# def normalise(mat, axis=1):
#     mat -= np.reshape(mat.mean(axis=axis), [mat.shape[0], 1])
#     mat /= np.reshape(mat.std(axis=axis), [mat.shape[0], 1])
#     mat /= np.sqrt(mat.shape[1])
#     return mat
#
# res_true_file = '/Users/damienarnol1/Documents/local/pro/PhD/social_effects/simulations/results_11/true_parameters.csv'
# res_simulated_file = \
#     '/Users/damienarnol1/Documents/local/pro/PhD/social_effects/simulations/results_11/simulated_parameters.csv'
#
# N_tests = 100
#
# results_true = np.zeros([N_tests, 3])
# results_simulated = np.zeros([N_tests, 3])
#
# for i in range(0, N_tests):
#
#     ########################################################################################
#     # simulating data
#     ########################################################################################
#     data_directory = '/Users/damienarnol1/Documents/local/pro/PhD/data/Zurich_data/output/tmp/'
#
#     X_file = data_directory + '/cell_locations_1'
#     phenotype_file = data_directory+'/membrane_1'
#
#     # read data
#     X = np.genfromtxt(X_file, delimiter=',', skiprows=1)
#     phenotypes = np.genfromtxt(phenotype_file, delimiter=' ', skiprows=1)
#     phenotypes = np.delete(phenotypes, 0, axis=1)
#     phenotypes = normalise(phenotypes)
#     N_cells = phenotypes.shape[0]
#
#     Kinship = phenotypes.dot(phenotypes.transpose())
#     Kinship -= np.linalg.eigvalsh(Kinship).min() * np.eye(N_cells)
#     Kinship *= covar_rescaling_factor(Kinship)
#     # Kinship =None
#     rm_diagonal = True
#     # possible values to switch off are local, noise, social, direct
#     switch_off = []
#     # switch_off = []
#
#     data, parameters = generate_data.generate_complete_model(N_cells, X, Kinship, switch_off, rm_diagonal)
#
#     ########################################################################################
#     # inference
#     ########################################################################################
#     # direct effect
#     direct_cov = FixedCov(parameters['Kinship'])
#
#     # noise
#     noise_cov = FixedCov(np.eye(N_cells))
#
#     # local_noise
#     local_noise_cov = SQExpCov(parameters['X'])
#     # local_noise_cov.length = parameters['local_noise_length']
#     local_noise_cov.length = N_cells / 6.0
#     # local_noise_cov.act_length = False
#
#     # environment effect
#     environment_cov = ZKZCov(parameters['X'], parameters['Kinship'], rm_diagonal)
#     # environment_cov.length = parameters['social_length']
#     environment_cov.length = N_cells / 6.0
#     # environment_cov.act_length = False
#
#     # total cov and mean
#     cov = SumCov(direct_cov, noise_cov)
#     cov = SumCov(cov, local_noise_cov)
#     cov = SumCov(cov, environment_cov)
#     # cov = SumCov(direct_cov, noise_cov)
#     # cov = SumCov(cov, local_noise_cov)
#
#     mean = MeanBase(data)
#
#     # define and optimise GP
#     gp = GP(covar=cov, mean=mean)
#
#     try:
#         gp.optimize()
#     except:
#         print('optimisation', str(i), 'failed')
#         continue
#
#     # show results
#     # print "true parameters vs inferred parameters "
#     # print "direct_scale = ", parameters['direct_scale'], " ", direct_cov.scale
#     # print "noise_scale = ", parameters['noise_scale'], " ", noise_cov.scale
#     # print "local_noise_scale = ", parameters['local_noise_scale'], " ", local_noise_cov.scale
#     # print "local_noise_length = ", parameters['local_noise_length'], " ", local_noise_cov.length
#     # print "environment_scale = ", parameters['social_scale'], " ", environment_cov.scale
#     # print "environment_length = ", parameters['social_length'], " ", environment_cov.length
#
#     results_true[i, :] = [parameters['social_scale'], parameters['direct_scale'],
#                           parameters['direct_scale']+parameters['noise_scale']+
#                           parameters['local_noise_scale']+parameters['social_scale']]
#     # results_simulated[i, :] = [direct_cov.scale, environment_cov.scale,
#     #                             environment_cov.scale + direct_cov.scale + noise_cov.scale + local_noise_cov.scale]
#     # results_simulated[i, :] = [direct_cov.scale, environment_cov.scale,
#     #                             environment_cov.scale + direct_cov.scale + noise_cov.scale]
#     # results_simulated[i, :] = [noise_cov.scale, direct_cov.scale,
#     #                             noise_cov.scale + direct_cov.scale + local_noise_cov.scale]
#     results_simulated[i, :] = [environment_cov.scale, direct_cov.scale,
#                                 noise_cov.scale + direct_cov.scale + local_noise_cov.scale + environment_cov.scale]
#     # print 'local noise scale ', local_noise_cov.scale
#
#     print('test', str(i), ' done ')
#
#
# np.savetxt(res_true_file, results_true, delimiter=' ')
# np.savetxt(res_simulated_file, results_simulated, delimiter=' ')
#
#
# #
# # N_cells = 1500
# # cov, se_params, Kinship = generate_data.generate_zkz(N_cells)
# # X = se_params[0]
# # scale = se_params[1]
# # length = se_params[2]
# # data = np.random.multivariate_normal(np.array([0] * N_cells), cov, 1).transpose()
# #
# # se = ZKZCov(X, Kinship)
# # # se.scale = scale
# # # se.length = length
# # gp = GP(MeanBase(data), se)
# # gp.optimize()
# #
# # print "true parameters vs inferred parameters "
# # print "scale = ", scale, " ", se.scale
# # print "length = ", length, " ", se.length
#
# #
# # N_cells = 1500
# # rm_zkz_diag = False
# #
# # data, parameters = generate_data.generate_simple(N_cells, rm_zkz_diag)
# #
# # # Kinship = parameters['Kinship']
# # phenotype = data
# # X = parameters['X']
# # length = parameters['length']
# # scale = parameters['scale']
# # noise_scale = parameters['noise']
# # Kinship = parameters['Kinship']
# # N = N_cells
# #
# # # defining ZKZ covariance term
# # zkz = ZKZCov(X, Kinship, rm_zkz_diag)
# # # zkz = SQExpCov(X)
# # # zkz.length = length
# # # zkz.scale = scale
# #
# # # fixed noise
# # noise = FixedCov(np.eye(N_cells))
# #
# # covar = SumCov(noise, zkz)
# # mean = MeanBase(phenotype)
# #
# # # optimise model
# # gp = GP(covar=covar, mean=mean)
# # # TODO initialise parameters
# # # TODO problem with calc_ste = True -> implement missing functions
# # gp.optimize(calc_ste=False)
# #
# # print "true parameters vs inferred parameters "
# # print "scale = ", scale, " ", zkz.scale
# # print "length = ", length, " ", zkz.length
# # print "noise scale = ", noise_scale, " ", noise.scale
# #
#
#
# # direct effect term
# # tmp = Kinship
# # tmp *= covar_rescaling_factor(Kinship)
# # directCov =FixedCov(tmp)
# #
# # # social effect term
# # tmp = Z.dot(Kinship.dot(Z.transpose()))
# # tmp *= covar_rescaling_factor(tmp)
# # socialCov = FixedCov(tmp)
# #
# # # noise term
# # noise = FixedCov(np.eye(N))
# #
# # # local Noise term
# # localNoise = SQExpCov(X)
# #
# # # total covariance
# # # TODO check syntax
# # covar = SumCov(directCov, socialCov)
# # covar = SumCov(covar, noise)
# # covar = SumCov(covar, localNoise)
# #
# # # mean
# # mean = phenotype
# #
# # gp = GP(covar=covar, mean=mean)
# # # TODO initialise parameters
# # gp.optimize(calc_ste=True)
# #
# #
# # # get results here



class cObject(object):
    """
    This class add to object a util function for caching
    """
    def __init__(self):
        pass

    def clear_cache(self,*method_names):
        for method_name in method_names:
            setattr(self,'_cache_'+method_name,None)
            setattr(self,'_cached_'+method_name,False)

    def fill_cache(self,method_name,value):
        setattr(self,'_cache_'+method_name,value)
        setattr(self,'_cached_'+method_name,True)

    def clear_cache_idxs(self,method_name,idxs):
        if type(idxs) is not tuple: idxs = [idxs]
        tail = '_'.join(str(idx) for idx in idxs)
        setattr(self,'_cache_'+method_name+'_'+tail,None)
        setattr(self,'_cached_'+method_name+'_'+tail,False)

    def fill_cache_idxs(self,method_name,idxs,value):
        if type(idxs) is not list: idxs = [idxs]
        tail = '_'.join(str(idx) for idx in idxs)
        setattr(self,'_cache_'+method_name+'_'+tail,value)
        setattr(self,'_cached_'+method_name+'_'+tail,True)

def cached(method):
    """ this function is used as a decorator for caching """
    _cache_attr_name = '_cache_'+method.__name__
    _bool_attr_name  = '_cached_'+method.__name__
    def method_wrapper(self,*args,**kwargs):
        is_cached = getattr(self,_bool_attr_name)
        if not is_cached:
            result = method(self, *args, **kwargs)
            setattr(self, _cache_attr_name, result)
            setattr(self, _bool_attr_name, True)
        return getattr(self,'_cache_'+method.__name__)
    return method_wrapper

def cached_idxs(method):
    """ this function is used as a decorator for caching """
    def method_wrapper(self,*args,**kwargs):
        tail = '_'.join(str(idx) for idx in args)
        _cache_attr_name = '_cache_'+method.__name__+'_'+tail
        _bool_attr_name  = '_cached_'+method.__name__+'_'+tail
        is_cached = getattr(self,_bool_attr_name)
        if not is_cached:
            result = method(self, *args, **kwargs)
            setattr(self, _cache_attr_name, result)
            setattr(self, _bool_attr_name, True)
        return getattr(self,_cache_attr_name)
    return method_wrapper




import sys
from limix.core.type.observed import Observed
from hcache import Cached, cached
from limix.core.utils.eigen import *
import scipy as SP
import pdb
import scipy.linalg as LA
import warnings

import logging as LG

class Covariance(Cached, Observed):
    """
    abstract super class for all implementations of covariance functions
    """
    def __init__(self,dim=None):
        Cached.__init__(self)
        if dim is not None:
            self.initialize(dim)

    def initialize(self,dim):
        self.dim = dim
        self._calcNumberParams()
        self._initParams()

    def clear_all(self):
        self.clear_cache('K','K_grad_i','logdet','logdet_grad_i',
                            'inv','chol','S','U',
                            'USi2','Sgrad','Ugrad')

    def setRandomParams(self):
        """
        set random hyperparameters
        """
        params = SP.randn(self.getNumberParams())
        self.setParams(params)

    def setCovariance(self,cov):
        """
        set hyperparameters from given covariance
        """
        warnings.warn('not implemented')

    def getParams(self):
        """
        get hyperparameters
        """
        return self.params

    def setParams(self,params):
        """
        set hyperParams
        """
        self.params = params
        self.clear_all()
        self._notify()

    def perturbParams(self,pertSize=1e-3):
        """
        slightly perturbs the values of the parameters
        """
        params = self.getParams()
        self.setParams(params+pertSize*SP.randn(params.shape[0]))

    def getNumberParams(self):
        """
        return the number of hyperparameters
        """
        return self.n_params

    def Kgrad_param(self,i):
        """
        partial derivative with repspect to the i-th hyperparamter theta[i]
        """
        LG.critical("implement Kgrad_theta")
        print(("%s: Function K not yet implemented"%(self.__class__)))

    def _calcNumberParams(self):
        """
        calculates the number of parameters
        """
        warnings.warn('not implemented')

    def _initParams(self):
         """
         initialize paramters to vector of zeros
         """
         params = SP.zeros(self.getNumberParams())
         self.setParams(params)

    def Kgrad_param_num(self,i,h=1e-4):
        """
        check discrepancies between numerical and analytical gradients
        """
        params = self.getParams()
        e = SP.zeros_like(params); e[i] = 1
        self.setParams(params-h*e)
        C_L = self.K()
        self.setParams(params+h*e)
        C_R = self.K()
        self.setParams(params)
        RV = (C_R-C_L)/(2*h)
        return RV

    ####################################
    # cached
    ####################################
    @cached
    def K(self):
        """
        evaluates the kernel for given hyperparameters theta and inputs X
        """
        LG.critical("implement K")
        print(("%s: Function K not yet implemented"%(self.__class__)))
        return None

    @cached
    def K_grad_i(self,i):
        """
        partial derivative with repspect to the i-th hyperparamter theta[i]
        """
        LG.critical("implement K_grad_i")
        print(("%s: Function K not yet implemented"%(self.__class__)))

    def solve(self,M):
        return LA.cho_solve((self.chol(),True),M)

    @cached
    def chol(self):
        return LA.cholesky(self.K()).T

    @cached
    def inv(self):
        return self.solve(SP.eye(self.dim))

    @cached
    def logdet(self):
        return 2*SP.log(SP.diag(self.chol())).sum()

    @cached
    def logdet_grad_i(self,i):
        return self.solve(self.K_grad_i(i)).diagonal().sum()

    @cached
    def S(self):
        RV,U = LA.eigh(self.K())
        self.fill_cache('U',U)
        return RV

    @cached
    def U(self):
        S,RV = LA.eigh(self.K())
        self.fill_cache('S',S)
        return RV

    @cached
    def USi2(self):
        # U * S**(-1/2)
        return self.U()*(self.S()**(-0.5))

    @cached
    def Sgrad(self,i):
        return dS_dti(self.Kgrad_param(i),U=self.U())

    @cached
    def Ugrad(self,i):
        return dU_dti(self.Kgrad_param(i),U=self.U(),S=self.S())

    @cached
    def USi2grad(self,i):
        # dU * S**(-1/2) + U * (-1/2 S**(-3/2) dS)
        Si2grad = -0.5*self.S()**(-1.5)*self.Sgrad(i)
        return self.Ugrad(i)*(self.S()**(-0.5)) + self.U()*Si2grad



__author__ = 'damienarnol1'

import pdb

import numpy as np

from limix.core.covar import SQExpCov
from limix.core.covar import ZKZCov
from limix.core.covar import FixedCov
from limix.core.covar import SumCov
from limix.core.gp import GP
from limix.core.mean import MeanBase
from limix.utils.preprocess import covar_rescaling_factor


def normalise(mat, axis=1):
    mat -= np.reshape(mat.mean(axis=axis), [mat.shape[0], 1])
    mat /= np.reshape(mat.std(axis=axis), [mat.shape[0], 1])
    mat /= np.sqrt(mat.shape[1])
    return mat

rm_diag = True

data_directory = '/Users/damienarnol1/Documents/local/pro/PhD/data/Zurich_data/output/tmp/'
output_dir = '/Users/damienarnol1/Documents/local/pro/PhD/social_effects/real_data/random_4/'

X_file = data_directory + '/all_positions'
phenotype_file = data_directory+'/membrane'

# read data
all_X = np.genfromtxt(X_file, delimiter=',', skiprows=1)
all_phenotypes = np.genfromtxt(phenotype_file, delimiter=' ', skiprows=1)

permutations = True


for image in range(0, 5):
    phenotypes = all_phenotypes[all_phenotypes[:, 0] == image, :]
    phenotypes = np.delete(phenotypes, 0, axis=1)
    phenotypes = normalise(phenotypes)

    X = all_X[all_X[:, 0] == image, :]
    X = np.delete(X, 0, axis=1)

    if permutations:
        X = X[np.random.permutation(X.shape[0]), :]

    if X.shape[0] != phenotypes.shape[0]:
        raise Exception('cell number inconsistent between position and epression levels ')

    N_cells = phenotypes.shape[0]

    parameters = np.zeros([phenotypes.shape[1], 6])

    output_file = output_dir + '_' + str(image)

    for phen in range(0, phenotypes.shape[1]):

        phenotype = phenotypes[:, phen]
        phenotype -= phenotype.mean()
        phenotype /= phenotype.std()
        phenotype = np.reshape(phenotype, [N_cells, 1])

        phenotypes_tmp = np.delete(phenotypes, phen, axis=1)

        Kinship = phenotypes_tmp.dot(phenotypes_tmp.transpose())
        Kinship -= np.linalg.eigvalsh(Kinship).min() * np.eye(N_cells)
        Kinship *= covar_rescaling_factor(Kinship)

        # create different models and print the result including likelihood
        # create all the covariance terms
        direct_cov = FixedCov(Kinship)

        # noise
        noise_cov = FixedCov(np.eye(N_cells))

        # local_noise
        local_noise_cov = SQExpCov(X)

        # environment effect
        environment_cov = ZKZCov(X, Kinship, rm_diag)

        # mean term
        mean = MeanBase(phenotype)

        #######################################################################
        # MODEL 1: all
        #######################################################################
        print('...........................................................')
        print('model 1 : complete model ')
        print('...........................................................')
        # total cov and mean
        cov = SumCov(direct_cov, noise_cov)
        cov = SumCov(cov, local_noise_cov)
        # direct_cov.scale = 0

        # local_noise_cov.scale = 0
        # direct_cov.scale = 0
        # cov = SumCov(noise_cov, local_noise_cov)

        cov = SumCov(cov, environment_cov)

        # fixing length scale of ZKZ and SE
        environment_cov.length = N_cells / 50
        # environment_cov.scale=0
        # environment_cov.act_length = False

        # local_noise_cov.length = N_cells/10.0
        # local_noise_cov.act_length = False

        # define and optimise GP
        gp = GP(covar=cov, mean=mean)

        try:
            gp.optimize()
        except:
            print('optimisation', str(phen), 'failed')
            continue

        # rescale each terms to sample variance one
        # direct cov: unnecessary as fixed covariance rescaled before optimisation
        # local noise covariance
        tmp = covar_rescaling_factor(local_noise_cov.K()/local_noise_cov.scale)
        local_noise_cov.scale /= tmp
        # env effect
        tmp = covar_rescaling_factor(environment_cov.K()/environment_cov.scale**2)
        environment_cov.scale = environment_cov.scale**2/tmp


        # show results
        print("inferred parameters ")
        print("direct_scale = ", " ", direct_cov.scale)
        print("noise_scale = ", " ", noise_cov.scale)
        print("local_noise_scale = ", " ", local_noise_cov.scale)
        print("local_noise_length = ", " ", local_noise_cov.length)
        print("environment_scale = ", " ", environment_cov.scale)
        print("environment_length = ", " ", environment_cov.length)

        parameters[phen, :] = [direct_cov.scale,
                               noise_cov.scale,
                               local_noise_cov.scale,
                               local_noise_cov.length,
                               environment_cov.scale,
                               environment_cov.length]

    result_header = 'direct_scale' + ' ' + \
                    'noise_scale' + ' ' + \
                    'local_noise_scale' + ' ' + \
                    'local_noise_length' + ' ' + \
                    'environment_scale' + ' ' + \
                    'environment_length'

    np.savetxt(output_file, parameters, delimiter=' ', header=result_header)





"""LMM testing code"""
import unittest
import scipy as sp
import numpy as np
import sys
from limix.core.covar import Cov2KronSumLR
from limix.core.covar import FreeFormCov
from limix.utils.check_grad import mcheck_grad
from limix.core.type.exception import TooExpensiveOperationError

class TestCov2KronSum(unittest.TestCase):
    def setUp(self):
        sp.random.seed(2)
        # define row caoriance
        n = 200
        f = 10
        P = 3
        X = 1.*(sp.rand(n, f)<0.2)
        # define col covariances
        Cn = FreeFormCov(P)
        self.C = Cov2KronSumLR(Cn = Cn, G = X, rank = 1)
        self.name = 'cov2kronSumLR'
        self.C.setRandomParams()

    def test_logdet_grad(self):
        def func(x, i):
            self.C.setParams(x)
            return self.C.logdet()

        def grad(x, i):
            self.C.setParams(x)
            return self.C.logdet_grad_i(i)

        x0 = self.C.getParams()
        err = mcheck_grad(func, grad, x0)
        np.testing.assert_almost_equal(err, 0., decimal = 3)

    def test_too_expensive_exceptions(self):
        n = 5001
        f = 10
        Cn = FreeFormCov(5001)
        X = 1.*(sp.rand(n, f)<0.2)
        C = Cov2KronSumLR(Cn=Cn, G=X, rank=1)
        with self.assertRaises(TooExpensiveOperationError):
            C.L()
        with self.assertRaises(TooExpensiveOperationError):
            C.R()
        with self.assertRaises(TooExpensiveOperationError):
            C.K()
        with self.assertRaises(TooExpensiveOperationError):
            C.K_grad_i(0)

    def test_param_activation(self):
        self.C.act_Cr = False
        self.C.act_Cn = False
        self.assertEqual(len(self.C.getParams()), 0)

        with self.assertRaises(ValueError):
            self.C.K_grad_i(0)

    def test_param_activation(self):
        self.C.act_Cr = False
        self.C.act_Cn = False
        self.assertEqual(len(self.C.getParams()), 0)

        with self.assertRaises(ValueError):
            self.C.K_grad_i(0)

    def test_inv(self):
        C = self.C

        L = sp.kron(C.Lc(), sp.eye(C.dim_r))
        W = sp.kron(C.Wc(), C.Wr())
        WdW = sp.dot(W.T, C.d()[:, sp.newaxis] * W)
        I_WdW = sp.eye(C.dim_c * C.dim_r) - WdW
        inv1 = sp.dot(L.T, sp.dot(I_WdW, L))

        inv2 = C.inv()
        np.testing.assert_array_almost_equal(inv1, inv2)

    def test_logdet(self):
        d1 = 2*np.log(np.diag(self.C.chol())).sum()
        d2 = self.C.logdet()
        np.testing.assert_almost_equal(d1, d2)

if __name__ == '__main__':
    unittest.main()



"""@package util
misc. utility functions used in limix modules and demos
"""

import numpy as np
import scipy as sp
import scipy as SP
import pdb, sys, pickle
import matplotlib.pylab as plt
import scipy.stats as st
import scipy.interpolate



def mean_impute(X, imissX=None, maxval=2.0):
    if imissX is None:
        imissX = np.isnan(X)
    
    n_i,n_s=X.shape
    if imissX is None:
        n_obs_SNP=np.ones(X.shape)
    else:    
        i_nonan=(~imissX)
        n_obs_SNP=i_nonan.sum(0)
        X[imissX]=0.0
    snp_sum=(X).sum(0)
    one_over_sqrt_pi=(1.0+snp_sum)/(2.0+maxval*n_obs_SNP)
    one_over_sqrt_pi=1./np.sqrt(one_over_sqrt_pi*(1.-one_over_sqrt_pi))
    snp_mean=(snp_sum*1.0)/(n_obs_SNP)

    X_ret=X-snp_mean
    X_ret*=one_over_sqrt_pi
    if imissX is not None:
        X_ret[imissX]=0.0
    return X_ret

def getPosNew(data):
    """
    get Fixed position
    """
    pos = data.geno['col_header']['pos'][:]
    chrom= data.geno['col_header']['chrom'][:]
    n_chroms = chrom.max()
    pos_new = []
    for chrom_i in range(1,n_chroms+1):
        I = chrom==chrom_i
        _pos = pos[I]
        for i in range(1,_pos.shape[0]):
            if not _pos[i]>_pos[i-1]:
                _pos[i:]=_pos[i:]+_pos[i-1]
        pos_new.append(_pos)
    pos_new = SP.concatenate(pos_new)
    return pos_new

def getCumPos(data):
    """
    getCumulativePosition
    """
    pos = getPosNew(data)
    chrom= data.geno['col_header']['chrom'][:]
    n_chroms = int(chrom.max())
    x = 0
    for chrom_i in range(1,n_chroms+1):
        I = chrom==chrom_i
        pos[I]+=x
        x=pos[I].max()
    return pos

def getChromBounds(data):
    """
    getChromBounds
    """
    chrom= data.geno['col_header']['chrom'][:]
    posCum = getCumPos(data)
    n_chroms = int(chrom.max())
    chrom_bounds = []
    for chrom_i in range(2,n_chroms+1):
        I1 = chrom==chrom_i
        I0 = chrom==chrom_i-1
        _cb = 0.5*(posCum[I0].max()+posCum[I1].min())
        chrom_bounds.append(_cb)
    return chrom_bounds